T_USERNAME = r'(^[a-z0-9]{3,16}$)'
T_PASSWORD = r'(^.{5,32}$)'
T_SESSIONID = r'(^[0-9a-f]{64}$)'
T_COMMANDID = r'(^[0-9a-f]{32}$)'
T_OBJECTNAME = r'(^.{1,63}$)'
T_COMMANDID_SIZE = 32
T_WORKER_SIZE = 32
T_PARAMETERS_SIZE = 1024
T_RESULT_SIZE = 4096
T_SESSIONID_SIZE = 64
T_USERNAME_SIZE = 16
T_PID = int



import re
import sys

ROUTES = []
WORKERS = []


def add_route(method, path):
    """
    Function decorator for HTTP method/path -> API function mapping.
    """
    def func_wrapper(function):
        global ROUTES
        splitpath = []
        elts = path.split('/')
        pos = 0
        root = None
        # Seeking for regexp in the path, because we need to compile it.
        for elt in elts:
            if len(elt) == 0:
                continue
            # regexp have to start with a ( end is terminated by a ).
            if elt[0] == '(' and elt[-1] == ')':
                # Append it compiled.
                splitpath.append(re.compile(elt))
            # string case
            else:
                if pos == 0:
                    root = elt
                splitpath.append(elt)
            pos += 1
        # A path can't start with a regexp.
        if root is None:
            raise Exception("Wrong route format.")
        ROUTES.append({
            'http_method': method,
            'root': root,
            'path': path,
            'splitpath': splitpath,
            'module': function.__module__,
            'function': function.__name__})
        return function
    return func_wrapper


def get_routes():
    """
    Returns the routes.
    """
    return ROUTES


def add_worker(name):
    """
    Function decorator for worker name -> worker function mapping.
    """
    def func_wrapper(function):
        global WORKERS
        WORKERS.append({
            'name': name,
            'module': function.__module__,
            'function': function.__name__})
        return function
    return func_wrapper


def get_worker(name):
    """
    Returns the function corresponding to the given worker name.
    """
    for worker in WORKERS:
        if worker['name'] == name:
            return getattr(sys.modules[worker['module']], worker['function'])
    raise Exception("Worker %s not found." % (name,))



import logging
import os.path
import re

from temboardagent.spc import pg_escape, error
from temboardagent.errors import HTTPError, NotificationError
from temboardagent.tools import validate_parameters
from temboardagent.notification import NotificationMgmt, Notification
from pgconf.types import (
    T_FILE_VERSION,
    T_NEW_VERSION,
    T_PGSETTINGS_FILTER,
)
from pgconf.hba import (
    HBAComment,
    HBAEntry,
    HBAManager,
)


logger = logging.getLogger(__name__)


def get_settings_categories(conn, config, _):
    query = """
SELECT DISTINCT(category) FROM pg_settings ORDER BY category
    """
    ret = {'categories': []}
    conn.execute(query)
    for row in conn.get_rows():
        ret['categories'].append(row['category'])
    return ret


def get_setting(conn, name):
    conn.execute("SELECT setting FROM pg_settings WHERE name = '%s'" % (name))
    return list(conn.get_rows())[0]['setting']


def get_settings(conn, config, http_context=None):
    auto_conf = get_auto_configuration(conn)
    file_conf = get_file_configuration(conn)
    filter_query = ''
    if http_context and 'filter' in http_context['query']:
        # Check 'filter' parameters.
        validate_parameters(http_context['query'], [
            ('filter', T_PGSETTINGS_FILTER, True)
        ])
        filter_query = " WHERE name ILIKE '%{0}%'" \
                       " OR short_desc ILIKE '%{0}%'" \
                       " OR extra_desc ILIKE '%{0}%'".format(
                               http_context['query']['filter'][0])
    query = """
SELECT
    name,
    setting,
    current_setting(name) AS current_setting,
    CASE WHEN name IN ('max_wal_size', 'min_wal_size')
    THEN current_setting('wal_segment_size') ELSE unit END AS unit,
    vartype,
    min_val,
    max_val,
    enumvals,
    context,
    category,
    short_desc||' '||coalesce(extra_desc, '') AS desc,
    boot_val
FROM pg_settings
%s ORDER BY category, name
    """ % (filter_query)
    conn.execute(query)
    ret = []
    do_not_format_names = ['unix_socket_permissions']
    for row in conn.get_rows():
        if http_context and len(http_context['urlvars']) > 0:
            if http_context['urlvars'][0] != row['category']:
                continue
        cat_exists = False
        i = 0
        for ret_cat in ret:
            if ret_cat['category'] == row['category']:
                cat_exists = True
                break
            i += 1
        in_auto_conf = False
        if row['name'] in auto_conf:
            in_auto_conf = True
        in_file_conf = False
        if row['name'] in file_conf:
            in_file_conf = True
        row_dict = {
            'name': row['name'],
            'setting': row['setting'],
            'setting_raw': row['current_setting'],
            'unit': row['unit'],
            'vartype': row['vartype'],
            'min_val': row['min_val'],
            'max_val': row['max_val'],
            'boot_val': row['boot_val'],
            'auto_val': None,
            'auto_val_raw': None,
            'file_val': None,
            'file_val_raw': None,
            'enumvals': row['enumvals'],
            'context': row['context'],
            'desc': row['desc']
        }

        if in_auto_conf:
            val = auto_conf[row['name']]
            if val.startswith("'") and val.endswith("'"):
                val = val[1:-1]
            if row_dict['vartype'] == u'bool':
                if val == 'true':
                    val = 'on'
                elif val == 'false':
                    val = 'off'
            row_dict['auto_val'] = val
            row_dict['auto_val_raw'] = val

        if in_file_conf:
            val = file_conf[row['name']]
            if val.startswith("'") and val.endswith("'"):
                val = val[1:-1]
            if row_dict['vartype'] == u'bool':
                if val == 'true':
                    val = 'on'
                elif val == 'false':
                    val = 'off'
            row_dict['file_val'] = val
            row_dict['file_val_raw'] = val

        if row['name'] not in do_not_format_names and \
           row['vartype'] == u'integer':
            row_dict['setting'] = int(row_dict['setting'])
            row_dict['min_val'] = int(row_dict['min_val'])
            row_dict['max_val'] = int(row_dict['max_val'])
            row_dict['boot_val'] = int(
                human_to_number(row_dict['boot_val'], row_dict['unit']))
            if in_auto_conf:
                row_dict['auto_val'] = int(
                    human_to_number(row_dict['auto_val'], row_dict['unit']))
            if in_file_conf:
                row_dict['file_val'] = int(
                    human_to_number(row_dict['file_val'], row_dict['unit']))
        elif (row['name'] not in do_not_format_names and
              row['vartype'] == u'real'):
            row_dict['setting'] = float(row_dict['setting'])
            row_dict['min_val'] = float(row_dict['min_val'])
            row_dict['max_val'] = float(row_dict['max_val'])
            row_dict['boot_val'] = float(row_dict['boot_val'])
            if in_auto_conf:
                row_dict['auto_val'] = float(row_dict['auto_val'])
            if in_file_conf:
                row_dict['file_val'] = float(row_dict['file_val'])

        if not cat_exists:
            ret.append({'category': row['category'],
                        'rows': [
                            row_dict
                        ]})
        else:
            ret[i]['rows'].append(row_dict)
    return ret


def human_to_number(h_value, h_unit=None):
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'YB', 'ZB']
    re_unit = re.compile(r'([0-9.]+)\s*([KMGBTPEYZ]?B)$', re.IGNORECASE)
    m_value = re_unit.match(str(h_value))
    factor = 1
    if h_unit:
        m_unit = re_unit.match(str(h_unit))
        if m_unit:
            factor = int(m_unit.group(1))
            h_unit = str(m_unit.group(2))

    if m_value:
        p_num = m_value.group(1)
        p_unit = m_value.group(2)
        m = 0
        for u in units:
            if h_unit and h_unit.lower() == u.lower():
                m = 0
            if u.lower() == p_unit.lower():
                return (int(p_num) * (1024 ** m))/factor
            else:
                m += 1

    # Valid time units are ms (milliseconds), s (seconds), min (minutes),
    # h (hours), and d (days
    re_unit = re.compile(r'([0-9.]+)\s*(ms|s|min|h|d)$')
    m_unit = re_unit.match(str(h_value))
    if h_unit == 'ms':
        mult = {'ms': 1, 's': 1000, 'min': 60000, 'h': 3600000, 'd': 86400000}
    elif h_unit == 's':
        mult = {'ms': -1000, 's': 1, 'min': 60, 'h': 3600, 'd': 86400}
    elif h_unit == 'min':
        mult = {'ms': -60000, 's': -60, 'min': 1, 'h': 60, 'd': 1440}
    elif h_unit == 'h':
        mult = {'ms': -3600000, 's': -3600, 'min': -60, 'h': 1, 'd': 24}
    elif h_unit == 'd':
        mult = {'ms': -86400000, 's': -86400, 'min': -1440, 'h': -24, 'd': 1}
    else:
        mult = {'ms': 1, 's': 1, 'min': 1, 'h': 1, 'd': 1}

    if m_unit:
        p_num = m_unit.group(1)
        p_unit = m_unit.group(2)
        if mult[p_unit] > 0:
            return (int(p_num) * mult[p_unit])
        else:
            return (int(p_num) / abs(mult[p_unit]))

    return h_value


def parse_configuration_file(file_path,):
    ret = {}
    if not os.path.isfile(file_path):
        return ret
    try:
        with open(file_path, 'r') as fp:
            file_content = fp.read()
        fp.close()
        reg_conf_entry = re.compile(r'^\s*([\w\.]+)\s*=\s*([^#]+)\s*(#.*)?$')
        for line in file_content.split('\n'):
            m_conf = reg_conf_entry.match(line.strip())
            if m_conf:
                ret[m_conf.group(1).strip()] = m_conf.group(2).strip()
        return ret
    except IOError:
        raise HTTPError(500, "Internal error.")


def get_auto_configuration(conn,):
    query = """
        SELECT setting AS data_dir
        FROM pg_settings
        WHERE name = 'data_directory'"""
    try:
        conn.execute(query)
        pg_data = list(conn.get_rows())[0]['data_dir']
        pg_auto_conf = "%s/postgresql.auto.conf" % (pg_data,)
        return parse_configuration_file(pg_auto_conf)
    except error:
        raise HTTPError(500, "Internal error.")


def get_file_configuration(conn,):
    query = """
        SELECT setting AS config_file
        FROM pg_settings
        WHERE name = 'config_file'"""
    try:
        conn.execute(query)
        config_file = list(conn.get_rows())[0]['config_file']
        return parse_configuration_file(config_file)
    except error:
        raise HTTPError(500, "Internal error.")


def get_settings_status(conn, config, http_context):
    settings = get_settings(conn, config)
    pending_restart_changes = []
    pending_reload_changes = []
    pending_reload = False
    pending_restart = False
    differ = False
    for category in settings:
        for row in category['rows']:
            differ = False
            if row['auto_val'] is not None and \
               row['auto_val'] != row['setting']:
                differ = True
                row['pending_val'] = row['auto_val_raw']
            if row['auto_val'] is None and row['file_val'] and \
               row['file_val'] != row['setting']:
                differ = True
                row['pending_val'] = row['file_val_raw']
            if differ:
                if row['context'] in \
                   ['backend', 'user', 'superuser', 'sighup']:
                    pending_reload = True
                    pending_reload_changes.append(row)
                elif row['context'] in ['postmaster']:
                    pending_restart = True
                    pending_restart_changes.append(row)
    return {
        'restart_pending': pending_restart,
        'reload_pending': pending_reload,
        'restart_changes': pending_restart_changes,
        'reload_changes': pending_reload_changes
    }


def post_settings(conn, config, http_context):
    if http_context and 'filter' in http_context['query']:
        # Check 'filter' parameters.
        validate_parameters(http_context['query'], [
            ('filter', T_PGSETTINGS_FILTER, True)
        ])
    pg_config_categories = get_settings(conn, config, None)
    if 'settings' not in http_context['post']:
        raise HTTPError(406, "Parameter 'settings' not sent.")
    settings = http_context['post']['settings']
    ret = {'settings': []}
    do_not_check_names = ['unix_socket_permissions', 'log_file_mode']
    logger.debug(settings)
    for setting in settings:
        if 'name' not in setting \
           or 'setting' not in setting:
            raise HTTPError(406, "setting item malformed.")
        checked = False
        try:
            for pg_config_category in pg_config_categories:
                for item in pg_config_category['rows']:
                    if item['name'] == setting['name']:
                        if item['name'] in do_not_check_names:
                            checked = True
                            raise Exception()
                        if item['vartype'] == u'integer':
                            # Integers handling.
                            if item['min_val'] and \
                               item['unit'] and \
                               (int(human_to_number(setting['setting'],
                                    item['unit'])) <
                                   int(item['min_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            if item['max_val'] and \
                               item['unit'] and \
                               (int(human_to_number(setting['setting'],
                                    item['unit'])) >
                                   int(item['max_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            setting['setting'] = pg_escape(setting['setting'])
                            if ((setting['setting'].startswith("'") and
                                 setting['setting'].endswith("'")) or
                                (setting['setting'].startswith('"') and
                                 setting['setting'].endswith('"'))):
                                setting['setting'] = setting['setting'][1:-1]
                            if setting['setting'] == '':
                                setting['setting'] = None
                            checked = True
                        if item['vartype'] == u'real':
                            # Real handling.
                            if item['min_val'] and \
                               (float(setting['setting']) <
                                   float(item['min_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            if item['max_val'] and \
                               (float(setting['setting']) >
                                   float(item['max_val'])):
                                raise HTTPError(406, "%s: Invalid setting." %
                                                     (item['name']))
                            setting['setting'] = float(setting['setting'])
                            checked = True
                        if item['vartype'] == u'bool':
                            # Boolean handling.
                            if setting['setting'].lower() not in \
                               [u'on', u'off']:
                                raise HTTPError(
                                    406, 'Invalid setting: %s.' %
                                         (setting['setting'].lower()))
                            checked = True
                        if item['vartype'] == u'enum':
                            # Enum handling.
                            if len(item['enumvals']) > 0:
                                enumvals = [
                                    re.sub(r"^[\"\'](.+)[\"\ ']$",
                                           r"\1", enumval)
                                    for enumval
                                    in item['enumvals'][1:-1].split(',')]
                                if ((setting['setting'].startswith("'") and
                                     setting['setting'].endswith("'")) or
                                    (setting['setting'].startswith('"') and
                                     setting['setting'].endswith('"'))):
                                    setting['setting'] = \
                                        setting['setting'][1:-1]
                                if setting['setting'] not in enumvals:
                                    raise HTTPError(
                                        406,
                                        'Invalid setting: %s.' %
                                        (setting['setting']))
                                checked = True
                        if item['vartype'] == u'string':
                            # String handling.
                            # setting must be escaped.
                            setting['setting'] = pg_escape(
                                str(setting['setting']))
                            if ((setting['setting'].startswith("'") and
                                 setting['setting'].endswith("'")) or
                                (setting['setting'].startswith('"') and
                                 setting['setting'].endswith('"'))):
                                setting['setting'] = setting['setting'][1:-1]
                            if setting['setting'] == '':
                                setting['setting'] = None
                            checked = True
                        raise Exception()
        except HTTPError as e:
            raise HTTPError(e.code, e.message['error'])
        except Exception as e:
            pass
        if not checked:
            raise HTTPError(406, 'Parameter %s can\'t be checked.' %
                                 (setting['name']))
        if 'force' not in setting:
            setting['force'] = 'false'
        if ((item['vartype'] == u'integer' and
            setting['setting'] != item['setting_raw']) or
            (item['vartype'] == u'real' and
            float(setting['setting']) != float(item['setting'])) or
            (item['vartype'] not in [u'integer', u'real'] and
            setting['setting'] != item['setting'])) or \
                (setting['force'] == 'true'):
            # At this point, all incoming parameters have been checked.
            if setting['setting']:
                query = "ALTER SYSTEM SET %s TO '%s'" % (setting['name'],
                                                         setting['setting'])
            else:
                query = "ALTER SYSTEM RESET %s;" % (setting['name'])

            logger.debug(query)

            # Push a notification on setting change.
            try:
                NotificationMgmt.push(
                    config,
                    Notification(
                        username=http_context['username'],
                        message="Setting '%s' changed: '%s' -> '%s'" % (
                            item['name'],
                            item['setting_raw'],
                            setting['setting'])))
            except NotificationError as e:
                logger.error(e.message)

            try:
                conn.execute(query)
            except error as e:
                raise HTTPError(408, "%s: %s" % (setting['name'], e.message))
            ret['settings'].append({
                'name': item['name'],
                'setting': setting['setting'],
                'previous_setting': item['setting_raw'],
                'restart': True if item['context'] in
                ['internal', 'postmaster'] else False
            })
    # Reload PG configuration.
    conn.execute("SELECT pg_reload_conf()")
    # Push a notification.
    try:
        NotificationMgmt.push(config, Notification(
                                username=http_context['username'],
                                message="PostgreSQL reload"))
    except NotificationError as e:
        logger.error(e.message)

    return ret


"""
HBA
"""


def get_hba_raw(conn, config, http_context):
    version = None

    if http_context and 'version' in http_context['query']:
        # Check parameter 'version'
        validate_parameters(http_context['query'], [
            ('version', T_FILE_VERSION, True)
        ])
        version = http_context['query']['version'][0]

    ret = {
        'filepath': None,
        'version': version,
        'content': ''
    }
    hba_file = get_setting(conn, 'hba_file')
    ret['filepath'] = hba_file
    ret['content'] = HBAManager.get_file_content(hba_file, version)
    return ret


def get_hba(conn, config, http_context):
    version = None
    if http_context and 'version' in http_context['query']:
        # Check parameter 'version'
        validate_parameters(http_context['query'], [
            ('version', T_FILE_VERSION, True)
        ])
        version = http_context['query']['version'][0]

    ret = {
        'filepath': None,
        'version': version,
        'entries': []
    }
    hba_file = get_setting(conn, 'hba_file')
    ret['filepath'] = hba_file
    for hba_entry in HBAManager.get_entries(hba_file, version):
        ret['entries'].append(hba_entry.__dict__)
    return ret


def get_hba_versions(conn, config, http_context):
    hba_file = get_setting(conn, 'hba_file')
    return {
        'filepath': hba_file,
        'versions': HBAManager.get_versions(hba_file)
    }


def post_hba_raw(conn, config, http_context):
    new_version = False

    if 'content' not in http_context['post']:
        raise HTTPError(406, "Parameter 'content' not sent.")
    if http_context and 'new_version' in http_context['post']:
        # Check parameter 'version'
        validate_parameters(http_context['post'], [
            ('new_version', T_NEW_VERSION, False)
        ])
        if http_context['post']['new_version'] is True:
            new_version = True

    hba_file = get_setting(conn, 'hba_file')
    return HBAManager.save_file_content(hba_file,
                                        http_context['post']['content'],
                                        new_version)


def post_hba(conn, config, http_context):
    new_version = False

    # Push a notification.
    try:
        NotificationMgmt.push(
            config,
            Notification(
                username=http_context['username'],
                message="HBA file updated"))
    except NotificationError as e:
        logger.error(e.message)

    if 'entries' not in http_context['post']:
        raise HTTPError(406, "Parameter 'entries' not sent.")

    if http_context and 'new_version' in http_context['post']:
        # Check parameter 'version'
        validate_parameters(http_context['post'], [
            ('new_version', T_NEW_VERSION, False)
        ])
        if http_context['post']['new_version'] is True:
            new_version = True

    hba_file = get_setting(conn, 'hba_file')
    hba_entries = []
    logger.debug(http_context['post']['entries'])
    for entry in http_context['post']['entries']:
        if 'comment' in entry and len(entry['connection']) == 0:
            new_hba_entry = HBAComment()
            new_hba_entry.comment = entry['comment']
        else:
            new_hba_entry = HBAEntry()
            try:
                new_hba_entry.connection = entry['connection'] \
                    if 'connection' in entry else ''
                new_hba_entry.database = entry['database'] \
                    if 'database' in entry else ''
                new_hba_entry.user = entry['user'] \
                    if 'user' in entry else ''
                new_hba_entry.address = entry['address'] \
                    if 'address' in entry else ''
                new_hba_entry.auth_method = entry['auth_method'] \
                    if 'auth_method' in entry else ''
                new_hba_entry.auth_options = entry['auth_options'] \
                    if 'auth_options' in entry else ''
            except Exception as e:
                logger.error(e.message)
                raise HTTPError(406, "Invalid HBA entry.")
            new_hba_entry.lazy_check()
        hba_entries.append(new_hba_entry)
    return HBAManager.save_entries(hba_file, hba_entries, new_version)


def delete_hba_version(conn, config, http_context):
    version = None

    if http_context and 'version' in http_context['query']:
        # Check parameter 'version'
        validate_parameters(http_context['query'], [
            ('version', T_FILE_VERSION, True)
        ])
        version = http_context['query']['version'][0]
    if version is None:
        raise HTTPError(406, "HBA version number must be specified.")

    hba_file = get_setting(conn, 'hba_file')
    # Push a notification.
    try:
        NotificationMgmt.push(
            config,
            Notification(
                username=http_context['username'],
                message="HBA file version '%s' removed." % (version)))
    except NotificationError as e:
        logger.error(e.message)

    return HBAManager.remove_version(hba_file, version)


def get_hba_options(conn, config, http_context):
    return HBAManager.options(conn)


"""
pg_ident
"""


def get_pg_ident(conn, config, http_context):
    ret = {
        'filepath': None,
        'content': ''
    }

    try:
        conn.execute("""
            SELECT setting
            FROM pg_settings
            WHERE name = 'ident_file'
        """)
        pg_ident_file = list(conn.get_rows())[0]['setting']
    except error as e:
        logger.error(str(e.message))
        raise HTTPError(500, 'Internal error.')

    ret['filepath'] = pg_ident_file
    with open(pg_ident_file, 'r') as fd:
        pg_ident_data = fd.read()
    fd.close()
    ret['content'] = pg_ident_data
    return ret


def post_pg_ident(conn, config, http_context):
    if 'content' not in http_context['post']:
        raise HTTPError(406, "Parameter 'content' not sent.")
    try:
        conn.execute("""
            SELECT setting
            FROM pg_settings
            WHERE name = 'ident_file'
        """)
        pg_ident_file = list(conn.get_rows())[0]['setting']
    except error as e:
        logger.error(str(e.message))
        raise HTTPError(500, 'Internal error.')

    with open(pg_ident_file, 'r') as fd:
        pg_ident_data = fd.read()
        fd.close()
        try:
            with open(pg_ident_file+".previous", 'w') as fdp:
                fdp.write(pg_ident_data)
                fdp.close()
        except Exception as e:
            raise HTTPError(500, 'Internal error.')

    with open(pg_ident_file, 'w') as fd:
        fd.write(http_context['post']['content'])
        fd.close()
    return {'update': True}



import logging
import re
import os
import time
import json

from temboardagent.spc import connector, error
from temboardagent.queue import Queue, Message
from temboardagent.tools import now
from temboardagent.inventory import SysInfo


def load_probes(options, home):
    """Give a list of probe objects, ready to run."""

    # All probes classes names start with "probe_", search for
    # classes and get an object
    probes = []
    r = re.compile(r'^probe_(\w+)$')
    for c in globals().keys():
        m = r.search(c)
        if m is not None \
           and issubclass(globals()[c], Probe) \
           and (m.group(1) in options['probes'] or '*' in options['probes']):
            o = eval(c + "(options)")
            o.set_home(home)
            probes.append(o)
            logging.info("Loaded probe: %s", o.get_name())

    return probes


def run_probes(probes, instances, delta=True):
    """Execute the probes."""

    logging.debug("Starting probe run")
    # Output is a mapping of probe names with lists. Each probe returns
    # a list of dicts(metric -> value).
    output = {}

    for p in probes:
        out = None
        if delta is False:
            p.delta_key = None
            p.delta_columns = None
            p.delta_interval_column = None

        if p.level == 'host':
            if p.check():
                logging.debug("Running host probe %s", p.get_name())
                out = p.run()

        if p.level == 'instance' or p.level == 'database':
            out = []
            for i in instances:
                if i['available']:
                    if p.check(i['version_num']):
                        logging.debug(
                            "Running %s level probe \"%s\" on instance \"%s\"",
                            p.level, p.get_name(), i['instance'])
                        out += p.run(i)

        output[p.get_name()] = out

    logging.debug("Finished probe run")
    return output


class Probe(object):
    """Base class for all plugins."""
    # At which level the information is gathered: host, instance or db
    level = None
    # Optionnal name of the probe
    name = None
    # Previous measures used for compute delta
    last_measure = {}
    last_measure_time = {}
    home = None

    def __init__(self, options):
        pass

    def check(self):
        """Check if the plugin can run on the target."""
        pass

    def run(self):
        """Returns the result."""
        pass

    def set_home(self, home):
        self.home = home

    def get_name(self):
        """Computes the name of the probe."""
        # Let the plugin overwrite the name
        if self.name is not None:
            return self.name

        # Compute the name from the class of the plugin
        m = re.search(r'^probe_(\w+)$', self.__class__.__name__.lower())
        if m is not None:
            return m.group(1)

        logging.error("Could not get the name of the probe")
        return None

    def get_last_measure(self, store_key):
        q = Queue("%s/%s.q" % (self.home, store_key), max_length=1,
                  overflow_mode='slide')
        msg = q.shift()
        if msg:
            return json.loads(msg.content)

    def delta(self, key, current_values):
        """
        Compute a delta between measures of two runs.

        Args:
            key (str): identify the values
            current_values (dict): mapping of latest measures

        Returns:
            a tuple of the time interval of the delta in seconds and a
            dict a delta with the same keys as the input.
        """
        current_time = time.time()
        store_key = self.get_name() + key
        last_measure = self.get_last_measure(store_key)
        delta = (None, None)
        delta_value = None
        # Compute deltas and update last_* variables
        try:
            if last_measure:
                delta_time = current_time - last_measure['measure_time']

                delta_values = {}
                for k in current_values.keys():
                    delta_value = current_values[k] - \
                        last_measure['measure'][k]
                    if delta_value < 0:
                        raise Exception('Negative delta value.')
                    delta_values[k] = delta_value

                delta = (delta_time, delta_values)
        except Exception as e:
            delta = (None, None)
        try:
            q = Queue("%s/%s.q" % (self.home, store_key), max_length=1,
                      overflow_mode='slide')
            q.push(Message(content=json.dumps({
                'measure_time': current_time,
                'measure': dict(current_values)})))
        except Exception as e:
            logging.error(str(e))
        return delta

    def __repr__(self):
        return self.get_name()


class HostProbe(Probe):
    """
        system probe base class
    """
    level = 'host'
    system = None  # kernel name from os.uname()[0]
    min_version = None
    max_version = None

    def check(self):
        """Check if the probe can run on this system."""
        if self.system is not None:
            if self.system != os.uname()[0]:
                return False

        version = [int(x) for x
                   in re.sub(r'[-_].*$', '', os.uname()[2]).split('.')]
        if self.min_version is not None:
            if version[0:len(self.min_version)] < self.min_version:
                return False

        if self.max_version is not None:
            if version[0:len(self.max_version)] > self.max_version:
                return False

        return True


class SqlProbe(Probe):
    """
        postgres probe base class
    """
    min_version = None
    max_version = None
    sql = None
    no_standby = False

    # Delta columns is a list of columns in the output of the query on
    # which delta are computed between to runs. When is it set, a new
    # key is added with the time interval of the delta named with
    # delta_interval_column. delta_key a column used as unique key to
    # compute delta on multiline output.
    delta_columns = None
    delta_key = None
    delta_interval_column = None

    def check(self, version=None):
        """Check if the plugin can run on the target version of PostgreSQL."""
        if version is None:
            return False

        if self.min_version is not None:
            if version < self.min_version:
                return False

        if self.max_version is not None:
            if version > self.max_version:
                return False

        return True

    def get_version(self, conninfo):
        conn = connector(conninfo['host'], conninfo['port'], conninfo['user'],
                         conninfo['password'], conninfo['database'])

        try:
            conn.connect()
            version = conn.get_pg_version()
            conn.close()
            return version
        except error:
            logging.error("Unable to get server version")

    def run_sql(self, conninfo, sql, database=None):
        """Get the result of the SQL query"""
        if sql is None:
            return []

        # Default the connection database to the one configured,
        # useful for instance level sql probes
        if database is None:
            database = conninfo['database']

        conn = connector(conninfo['host'], conninfo['port'], conninfo['user'],
                         conninfo['password'], database)

        output = []
        try:
            conn.connect()
            conn.execute(sql)
            for r in conn.get_rows():
                # Add the info of the instance (port) to the
                # result to output one big list for all instances and
                # all databases
                r['port'] = conninfo['port']

                # Compute delta if the probe needs that
                if self.delta_columns is not None:
                    to_delta = {}

                    # XXX. Convert results to float(), spc retrieves
                    # everything as string. So far psycopg2 on the
                    # server side handles to rest
                    for k in self.delta_columns:
                        if k in r.keys():
                            to_delta[k] = float(r[k])

                    # Create the store key for the delta
                    if self.delta_key is not None:
                        key = conninfo['instance'] + database + \
                            r[self.delta_key]
                    else:
                        key = conninfo['instance'] + database

                    # Calculate delta
                    (interval, deltas) = self.delta(key, to_delta)

                    # The first time, no delta is returned
                    if interval is None:
                        continue

                    # Merge result and add the interval column
                    r.update(deltas)
                    r[self.delta_interval_column] = interval

                output.append(r)
            conn.close()
        except error:
            logging.error(
                "Unable to run probe \"%s\" on \"%s\" on database \"%s\"",
                self.get_name(), conninfo['instance'], database)
        return output

    def run(self, conninfo):
        """Execute the query depending on the level configured."""
        if self.no_standby and conninfo['standby']:
            return []

        if self.level == 'instance':
            return self.run_sql(conninfo, self.sql)

        if self.level == 'database':
            output = []
            for db in conninfo['dbnames']:
                output += self.run_sql(conninfo, self.sql, db['dbname'])
            return output


class probe_sessions(SqlProbe):
    level = 'instance'

    def check(self, version):
        if not super(probe_sessions, self).check(version):
            return False

        if version < 90200:
            self.sql = """select
  current_timestamp as datetime,
  d.datname as dbname,
  coalesce(sum((current_query not in ('<IDLE>','<IDLE> in transaction') and not waiting)::integer), 0) as active,
  coalesce(sum(waiting::integer), 0) as waiting,
  coalesce(sum((current_query='<IDLE>')::integer), 0) as idle,
  coalesce(sum((current_query='<IDLE> in transaction')::integer), 0) as idle_in_xact,
  0 as idle_in_xact_aborted,
  coalesce(sum((current_query='<FASTPATH> function call')::integer), 0) as fastpath,
  coalesce(sum((current_query='<command string not enabled>')::integer), 0) as disabled,
  coalesce(sum((current_query='<insufficient privilege>')::integer), 0) as no_priv
from pg_database d
  left join pg_stat_activity a on (d.oid = a.datid)
where d.datallowconn
group by d.datname"""  # noqa
        elif version >= 90200 and version < 90600:
            self.sql = """select
  current_timestamp as datetime,
  d.datname as dbname,
  coalesce(sum((state = 'active' and not waiting)::integer), 0) as active,
  coalesce(sum((state = 'active' and waiting)::integer), 0) as waiting,
  coalesce(sum((state = 'idle')::integer), 0) as idle,
  coalesce(sum((state = 'idle in transaction')::integer), 0) as idle_in_xact,
  coalesce(sum((state = 'idle in transaction (aborted)')::integer), 0) as idle_in_xact_aborted,
  coalesce(sum((state = 'fastpath function call')::integer), 0) as fastpath,
  coalesce(sum((state = 'disabled')::integer), 0) as disabled,
  coalesce(sum((query = '<insufficient privilege>')::integer), 0) as no_priv
from pg_database d
  left join pg_stat_activity a on (d.oid = a.datid)
where d.datallowconn
group by d.datname"""  # noqa
        elif version >= 90600:
            self.sql = """select
  current_timestamp as datetime,
  d.datname as dbname,
  coalesce(sum((state = 'active' and wait_event_type IS DISTINCT FROM 'Lock')::integer), 0) as active,
  coalesce(sum((state = 'active' and wait_event_type IS NOT DISTINCT FROM 'Lock')::integer), 0) as waiting,
  coalesce(sum((state = 'idle')::integer), 0) as idle,
  coalesce(sum((state = 'idle in transaction')::integer), 0) as idle_in_xact,
  coalesce(sum((state = 'idle in transaction (aborted)')::integer), 0) as idle_in_xact_aborted,
  coalesce(sum((state = 'fastpath function call')::integer), 0) as fastpath,
  coalesce(sum((state = 'disabled')::integer), 0) as disabled,
  coalesce(sum((query = '<insufficient privilege>')::integer), 0) as no_priv
from pg_database d
  left join pg_stat_activity a on (d.oid = a.datid)
where d.datallowconn
group by d.datname"""  # noqa

        return True


class probe_xacts(SqlProbe):
    level = 'instance'
    min_version = 70400
    sql = """select
  current_timestamp as datetime,
  s.datname as dbname,
  xact_commit as n_commit,
  xact_rollback as n_rollback
from pg_stat_database s
  join pg_database d on (d.oid = s.datid)
where d.datallowconn"""
    delta_columns = ['n_commit', 'n_rollback']
    delta_key = 'dbname'
    delta_interval_column = 'measure_interval'


class probe_locks(SqlProbe):
    level = 'instance'
    sql = """select
  current_timestamp as datetime,
  d.datname as dbname,
  coalesce(sum((mode = 'AccessShareLock' and granted)::integer), 0) as access_share,
  coalesce(sum((mode = 'AccessShareLock' and not granted)::integer), 0) as waiting_access_share,
  coalesce(sum((mode = 'RowShareLock' and granted)::integer), 0) as row_share,
  coalesce(sum((mode = 'RowShareLock' and not granted)::integer), 0) as waiting_row_share,
  coalesce(sum((mode = 'RowExclusiveLock' and granted)::integer), 0) as row_exclusive,
  coalesce(sum((mode = 'RowExclusiveLock' and not granted)::integer), 0) as waiting_row_exclusive,
  coalesce(sum((mode = 'ShareUpdateExclusiveLock' and granted)::integer), 0) as share_update_exclusive,
  coalesce(sum((mode = 'ShareUpdateExclusiveLock' and not granted)::integer), 0) as waiting_share_update_exclusive,
  coalesce(sum((mode = 'ShareLock' and granted)::integer), 0) as share,
  coalesce(sum((mode = 'ShareLock' and not granted)::integer), 0) as waiting_share,
  coalesce(sum((mode = 'ShareRowExclusiveLock' and granted)::integer), 0) as share_row_exclusive,
  coalesce(sum((mode = 'ShareRowExclusiveLock' and not granted)::integer), 0) as waiting_share_row_exclusive,
  coalesce(sum((mode = 'ExclusiveLock' and granted)::integer), 0) as exclusive,
  coalesce(sum((mode = 'ExclusiveLock' and not granted)::integer), 0) as waiting_exclusive,
  coalesce(sum((mode = 'AccessExclusiveLock' and granted)::integer), 0) as access_exclusive,
  coalesce(sum((mode = 'AccessExclusiveLock' and not granted)::integer), 0) as waiting_access_exclusive,
  coalesce(sum((mode = 'SIReadLock' and granted)::integer), 0) as siread
from pg_database d
  left join pg_locks l on (l.database = d.oid)
where d.datallowconn
group by d.datname"""  # noqa


class probe_blocks(SqlProbe):
    level = 'instance'
    sql = """select
  current_timestamp as datetime,
  s.datname as dbname,
  blks_read,
  blks_hit,
  coalesce(blks_hit::float*100/nullif(blks_read+blks_hit, 0), 0) as hitmiss_ratio
from pg_stat_database s
  join pg_database d on (d.oid = s.datid)
where d.datallowconn"""  # noqa
    delta_columns = ['blks_read', 'blks_hit']
    delta_key = 'dbname'
    delta_interval_column = 'measure_interval'


class probe_bgwriter(SqlProbe):
    level = 'instance'
    min_version = 80300
    sql = """select current_timestamp as datetime, * from pg_stat_bgwriter"""
    delta_columns = [
        'checkpoints_timed', 'checkpoints_req', 'checkpoint_write_time',
        'checkpoint_sync_time', 'buffers_checkpoint', 'buffers_clean',
        'maxwritten_clean', 'buffers_backend', 'buffers_backend_fsync',
        'buffers_alloc'
    ]
    delta_interval_column = 'measure_interval'


class probe_db_size(SqlProbe):
    level = 'instance'
    sql = """select
  current_timestamp as datetime,
  datname as dbname,
  pg_database_size(oid) as size
from pg_database
where datallowconn"""


class probe_tblspc_size(SqlProbe):
    level = 'instance'
    sql = """select
  current_timestamp as datetime,
  spcname,
  pg_tablespace_size(oid) as size
from pg_tablespace"""


class probe_filesystems_size(HostProbe):
    system = 'Linux'

    def run(self):
        # Everything is already gathered in the inventory, just add
        # the time
        out = []
        datetime = now()
        sysinfo = SysInfo()
        for fs in sysinfo.file_systems():
            fs['datetime'] = datetime
            out.append(fs)

        return out


class probe_cpu(HostProbe):
    system = 'Linux'
    hz = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

    def run(self):
        stat = open('/proc/stat')
        out = []
        for line in stat:
            cols = line.split()
            if len(cols) == 0:
                continue
            if re.match(r'^cpu\d+$', cols[0]):
                # Convert values to int then in milliseconds,
                # before computing the delta
                (interval, metrics) = self.delta(cols[0], {
                    'time_user': (int(cols[1]) + int(cols[2]))
                    * 1000 / self.hz,
                    'time_system': (int(cols[3]) + int(cols[6]) + int(cols[7]))
                    * 1000 / self.hz,
                    'time_idle': int(cols[4]) * 1000 / self.hz,
                    'time_iowait': int(cols[5]) * 1000 / self.hz,
                    'time_steal': int(cols[8]) * 1000 / self.hz
                })

                if interval is None:
                    continue

                metrics['measure_interval'] = interval
                metrics['datetime'] = now()
                metrics['cpu'] = cols[0]

                out.append(metrics)

        stat.close()

        return out


class probe_process(HostProbe):
    system = 'Linux'

    def run(self):
        metrics = {
            'datetime': now()
        }
        # Process information is partly stored in /proc/stat, ctxt and
        # processes are ever incresing counters, compute deltas on
        # them.
        to_delta = {}
        stat = open('/proc/stat')
        for line in stat:
            cols = line.split()
            if len(cols) == 0:
                continue
            if cols[0] == 'ctxt':
                to_delta['context_switches'] = int(cols[1])
            if cols[0] == 'processes':
                to_delta['forks'] = int(cols[1])
            if cols[0] == 'procs_running':
                metrics['procs_running'] = int(cols[1])
            if cols[0] == 'procs_blocked':
                metrics['procs_blocked'] = int(cols[1])
        stat.close()

        # Total number of process is stored in /proc/loadavg
        load = open('/proc/loadavg')
        m = re.match(r'.*/(\d+) ', load.readline())
        load.close()
        if m:
            metrics['procs_total'] = m.group(1)
        else:
            metrics['procs_total'] = 0

        # Compute deltas for values of /proc/stat since boot time
        (interval, deltas) = self.delta('key', to_delta)
        # No deltas on the first call
        if interval is None:
            return []

        metrics['measure_interval'] = interval
        metrics.update(deltas)
        return [metrics]


class probe_memory(HostProbe):
    system = 'Linux'

    def run(self):
        sysinfo = SysInfo()
        meminfo = sysinfo.mem_info()

        return [{
            'datetime': now(),
            'mem_total': meminfo['MemTotal'],
            'mem_used': meminfo['MemTotal'] - meminfo['MemFree'],
            'mem_free': meminfo['MemFree'],
            'mem_buffers': meminfo['Buffers'],
            'mem_cached': meminfo['Cached'],
            'swap_total': meminfo['SwapTotal'],
            'swap_used': meminfo['SwapTotal'] - meminfo['SwapFree']
        }]


class probe_loadavg(HostProbe):
    system = 'Linux'

    def run(self):
        loadavg = open('/proc/loadavg')
        cols = loadavg.readline().split()
        loadavg.close()
        return [{
            'datetime': now(),
            'load1': cols[0],
            'load5': cols[1],
            'load15': cols[2]
        }]


class probe_wal_files(SqlProbe):
    level = 'instance'
    min_version = 80200

    def run(self, conninfo):
        version = self.get_version(conninfo)

        if conninfo['standby']:
            return []

        metric = {
            'datetime': now(),
            'port': conninfo['port']
        }
        if version < 100000:
            sql = """
            SELECT count(s.f) AS total,
                   sum((pg_stat_file('pg_xlog/'||s.f)).size) AS total_size,
                   pg_current_xlog_location() as current_location
            FROM pg_ls_dir('pg_xlog') AS s(f)
            WHERE f ~ E'^[0-9A-F]{24}$'
            """
        else:
            sql = """
            SELECT count(s.f) AS total,
                   sum((pg_stat_file('pg_wal/'||s.f)).size) AS total_size,
                   pg_current_wal_lsn() as current_location
            FROM pg_ls_dir('pg_wal') AS s(f)
            WHERE f ~ E'^[0-9A-F]{24}$'
            """
        rows = self.run_sql(conninfo, sql)

        metric['total'] = rows[0]['total']
        metric['total_size'] = rows[0]['total_size']
        metric['current_location'] = rows[0]['current_location']

        if version < 100000:
            sql = """
            SELECT count(s.f) AS archive_ready
            FROM pg_ls_dir('pg_xlog/archive_status') AS s(f)
            WHERE f ~ E'\.ready$'
            """
        else:
            sql = """
            SELECT count(s.f) AS archive_ready
            FROM pg_ls_dir('pg_wal/archive_status') AS s(f)
            WHERE f ~ E'\.ready$'
            """
        rows = self.run_sql(conninfo, sql)

        metric['archive_ready'] = rows[0]['archive_ready']

        # Calcutate the written size by using the delta between the
        # position between to runs. The current xlog location must be
        # converted to an number first
        m = re.match(r'^([0-9A-F]+)/([0-9A-F]+)$', metric['current_location'])
        if m:
            current = int("0xff000000", 0) * \
                int("0x" + m.group(1), 0) + int("0x" + m.group(2), 0)
        else:
            logging.error("Unable to convert xlog location to a number")
            return []

        (interval, delta) = self.delta(conninfo['instance'],
                                       {'written_size': current})

        # Empty the first time
        if interval is None:
            return []

        metric['measure_interval'] = interval
        metric.update(delta)

        return [metric]


class probe_replication(SqlProbe):
    level = 'instance'
    min_version = 90000

    def run(self, conninfo):
        if conninfo['standby']:
            return self.run_sql(conninfo, """select
  current_timestamp as datetime,
  pg_last_xlog_receive_location() as receive_location,
  pg_last_xlog_replay_location() as replay_location""")
        else:
            return []



from argparse import ArgumentParser, SUPPRESS as UNDEFINED_ARGUMENT
from socket import getfqdn
import logging
from multiprocessing import Process, Queue
import os
import signal

from ..cli import cli, define_common_arguments
from ..sharedmemory import Commands, Sessions
from ..async import Scheduler
from ..configuration import load_configuration, OptionSpec
from ..daemon import (
    daemonize,
    httpd_sigterm_handler,
    set_global_scheduler,
    httpd_sighup_handler,
)
from ..httpd import httpd_run
from ..queue import purge_queue_dir
from .. import validators as v


logger = logging.getLogger(__name__)


def define_arguments(parser):
    define_common_arguments(parser)
    parser.add_argument(
        '-d', '--daemon',
        action='store_true', dest='temboard_daemonize',
        help="Run in background.",
    )
    parser.add_argument(
        '-p', '--pid-file',
        action='store', dest='temboard_pidfile',
        help="PID file.",
    )


def list_options_specs():
    # Generate each option specs.
    section = 'temboard'
    yield OptionSpec(section, 'daemonize', default=False)
    yield OptionSpec(section, 'pidfile', default='/run/temboard-agent.pid')
    yield OptionSpec(
        section, 'address', default='0.0.0.0', validator=v.address)
    yield OptionSpec(section, 'port', validator=v.port, default=2345)
    yield OptionSpec(
        section, 'ssl_cert_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(
        section, 'ssl_key_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(section, 'ssl_ca_cert_file', validator=v.file_)
    yield OptionSpec(section, 'key')
    yield OptionSpec(
        section, 'users',
        default='/etc/temboard-agent/users', validator=v.file_,
    )
    yield OptionSpec(section, 'hostname', default=getfqdn())
    home = os.environ.get('HOME', '/var/lib/temboard-agent')
    yield OptionSpec(section, 'home', default=home, validator=v.writeabledir)
    all_plugins = [
        "activity",
        "administration",
        "dashboard",
        "monitoring",
        "pgconf",
    ]
    yield OptionSpec(
        section, 'plugins', default=all_plugins, validator=v.jsonlist,
    )

    s = 'postgresql'
    yield OptionSpec(
        s, 'host', default='/var/run/postgresql', validator=v.dir_)
    yield OptionSpec(s, 'instance', default='main')
    yield OptionSpec(s, 'port', default=5432, validator=v.port)
    yield OptionSpec(s, 'user', default='postgres')
    yield OptionSpec(s, 'password')
    yield OptionSpec(s, 'dbname', default='postgres')

    s = 'logging'
    yield OptionSpec(s, 'method', default='syslog', validator=v.logmethod)
    yield OptionSpec(s, 'level', default='INFO', validator=v.loglevel)
    yield OptionSpec(
        s, 'facility', default='local0', validator=v.syslogfacility,
    )
    yield OptionSpec(s, 'destination', default='/dev/log')


@cli
def main(argv, environ):
    parser = ArgumentParser(
        prog='temboard-agent',
        description="temBoard agent.",
        argument_default=UNDEFINED_ARGUMENT,
    )
    define_arguments(parser)
    args = parser.parse_args(argv)
    config = load_configuration(
        specs=list_options_specs(),
        args=args, environ=environ,
    )

    # Run temboard-agent as a background daemon.
    if (config.temboard.daemonize):
        daemonize(config.temboard.pidfile)

    config.setup_logging()
    logger.info("Starting main process.")

    # Purge all data queues at start time excepting metrics & notifications.
    purge_queue_dir(config.temboard['home'],
                    ['metrics.q', 'notifications.q', 'notifications_last_10.q']
                    )

    # Creation of the command list (max 100).
    commands = Commands(100)
    # Creation of the session list (max 100).
    sessions = Sessions(100)
    # Command queue creation.
    queue_in = Queue()

    # Start the command scheduler process.
    scheduler = Process(target=Scheduler,
                        args=(commands, queue_in, config, sessions))
    scheduler.start()

    # Let's store scheduler reference in a global var.
    set_global_scheduler(scheduler)
    # Add signal handlers on SIGTERM and SIGHUP.
    signal.signal(signal.SIGTERM, httpd_sigterm_handler)
    signal.signal(signal.SIGHUP, httpd_sighup_handler)

    # Serve HTTPS forever.
    httpd_run(commands, queue_in, config, sessions)

    # Join command scheduler process on http server process exit.
    scheduler.join()

    return 0


if __name__ == '__main__':  # pragma: no cover
    main()



import logging
import time
import pickle
import base64
import os
try:
    from configparser import NoOptionError
except ImportError:
    from ConfigParser import NoOptionError

from temboardagent.routing import add_route, add_worker
from temboardagent.api_wrapper import api_function_wrapper_pg
from temboardagent.configuration import (
    PluginConfiguration,
    ConfigurationError,
)
from temboardagent.api import check_sessionid
from temboardagent.tools import validate_parameters, hash_id
from temboardagent.types import T_OBJECTNAME
from temboardagent.sharedmemory import Command
from temboardagent.errors import (
    HTTPError,
    SharedItem_exists,
    SharedItem_no_free_slot_left,
    SharedItem_not_found,
    NotificationError,
)
from temboardagent.workers import COMMAND_START, COMMAND_DONE, COMMAND_ERROR
from temboardagent.spc import connector, error
from temboardagent.command import (
    oneline_cmd_to_array,
    exec_script,
)
from temboardagent.notification import NotificationMgmt, Notification

import administration.functions as admin_functions
from administration.types import T_CONTROL, T_VACUUMMODE


logger = logging.getLogger(__name__)


@add_route('GET', '/administration/pg_version')
def api_pg_version(http_context,
                   queue_in=None,
                   config=None,
                   sessions=None,
                   commands=None):
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   admin_functions,
                                   'pg_version')


@add_route('POST', '/administration/control')
def post_pg_control(http_context,
                    queue_in=None,
                    config=None,
                    sessions=None,
                    commands=None):
    """
Control PostgreSQL server. Supported actions are "start", "stop", "restart" and "reload".

**Example request**:

.. sourcecode:: http

    POST /administration/control HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e
    Content-Type: application/json

    {
        "action": "restart"
    }

**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:57:52 GMT
    Content-type: application/json

    {
        "action": "restart",
        "state": "ok"
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header or parameter is malformed.

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Parameter 'action' is malformed."}


    """  # noqa
    # NOTE: in this case we don't want to use api functions wrapper, it leads
    # to "Broken pipe" error with debian init.d scrip on start/restart. This is
    # probably due to getattr() call.
    post = http_context['post']

    try:
        check_sessionid(http_context['headers'], sessions)
        # Check POST parameters.
        validate_parameters(post, [
            ('action', T_CONTROL, False)
        ])
        session = sessions.get_by_sessionid(
                    http_context['headers']['X-Session'].encode('utf-8')
                    )
    except (Exception, HTTPError) as e:
        logger.exception(str(e))
        logger.debug(http_context)
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")

    try:
        NotificationMgmt.push(config,
                              Notification(
                                username=session.username,
                                message="PostgreSQL %s" % post['action']
                                )
                              )
    except (NotificationError, Exception) as e:
        logger.exception(str(e))

    try:
        logger.info("PostgreSQL '%s' requested." % (post['action']))
        cmd_args = oneline_cmd_to_array(
                    config.plugins['administration']['pg_ctl'] % (
                        post['action']
                        )
                    )
        (rcode, stdout, stderr) = exec_script(cmd_args)
        if rcode != 0:
            raise Exception(str(stderr))
        # Let's check if PostgreSQL is up & running after having executed
        # 'start' or 'restart' action.
        if post['action'] in ['start', 'restart']:
            conn = connector(
                host=config.postgresql['host'],
                port=config.postgresql['port'],
                user=config.postgresql['user'],
                password=config.postgresql['password'],
                database=config.postgresql['dbname']
            )
            # When a start/restart operation is requested, after the
            # startup/pg_ctl script has been executed then we check that
            # postgres is up & running:
            # while the PG conn. is not working then, for 10 seconds (max)
            # we'll check (connect/SELECT 1/disconnect) the connection, every
            # 0.5 second.
            retry = True
            t_start = time.time()
            while retry:
                try:
                    conn.connect()
                    conn.execute('SELECT 1')
                    conn.close()
                    logger.info("Done.")
                    return {'action': post['action'], 'state': 'ok'}
                except error:
                    if (time.time() - t_start) > 10:
                        try:
                            conn.close()
                        except error:
                            pass
                        except Exception:
                            pass
                        logger.info("Failed.")
                        return {'action': post['action'], 'state': 'ko'}
                time.sleep(0.5)

        elif post['action'] == 'stop':
            conn = connector(
                host=config.postgresql['host'],
                port=config.postgresql['port'],
                user=config.postgresql['user'],
                password=config.postgresql['password'],
                database=config.postgresql['dbname']
            )
            # Check the PG conn is not working anymore.
            try:
                retry = True
                t_start = time.time()
                while retry:
                    conn.connect()
                    conn.execute('SELECT 1')
                    conn.close()
                    time.sleep(0.5)
                    if (time.time() - t_start) > 10:
                        retry = False
                logger.info("Failed.")
                return {'action': post['action'], 'state': 'ko'}
            except error:
                logger.info("Done.")
                return {'action': post['action'], 'state': 'ok'}
        logger.info("Done.")
        return {'action': post['action'], 'state': 'ok'}
    except (Exception, error, HTTPError) as e:
        logger.exception(str(e))
        logger.info("Failed")
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")


@add_route('POST', '/administration/vacuum')
def api_vacuum(http_context,
               queue_in=None,
               config=None,
               sessions=None,
               commands=None):
    worker = b'vacuum'
    try:
        check_sessionid(http_context['headers'], sessions)
        post = http_context['post']
        # Check POST parameters.
        validate_parameters(post, [
            ('database', T_OBJECTNAME, False),
            ('table', T_OBJECTNAME, False),
            ('mode', T_VACUUMMODE, False)
        ])
        # Serialize parameters.
        parameters = base64.b64encode(
                        pickle.dumps(
                            {
                                'database': post['database'],
                                'table': post['table'],
                                'mode': post['mode']
                            }
                        )
                    ).decode('utf-8')
    except (Exception, HTTPError) as e:
        logger.exception(str(e))
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")

    # Check command uniqueness.
    try:
        commands.check_uniqueness(worker, parameters)
    except SharedItem_exists as e:
        logger.exception(str(e))
        raise HTTPError(402,
                        "Vaccum '%s' already running on table '%s'." % (
                            post['mode'],
                            post['table']
                            )
                        )
    cid = hash_id(worker + b'-' + parameters.encode('utf-8'))
    command = Command(
            cid.encode('utf-8'),
            time.time(),
            0,
            worker,
            parameters,
            0,
            u'')
    try:
        commands.add(command)
        # Put the Command in the command queue
        queue_in.put(command)
        return {"cid": cid}
    except SharedItem_no_free_slot_left as e:
        logger.exception(str(e))
        raise HTTPError(500, "Internal error.")


@add_worker(b'vacuum')
def worker_vacuum(commands, command, config):
    start_time = time.time() * 1000
    logger.info("Starting with pid=%s" % (os.getpid()))
    logger.debug("commandid=%s" % (command.commandid,))

    try:
        command.state = COMMAND_START
        command.time = time.time()
        commands.update(command)
        parameters = pickle.loads(base64.b64decode(command.parameters))
        logger.debug("table=%s, mode=%s, database=%s" % (
                        parameters['table'],
                        parameters['mode'],
                        parameters['database'],
                        )
                     )

        conn = connector(
            host=config.postgresql['host'],
            port=config.postgresql['port'],
            user=config.postgresql['user'],
            password=config.postgresql['password'],
            database=parameters['database']
        )
        conn.connect()
        if parameters['mode'] == 'standard':
            query = "VACUUM %s" % (parameters['table'],)
        else:
            query = "VACUUM %s %s" % (parameters['mode'], parameters['table'],)
        conn.execute(query)
        conn.close()
    except (error, SharedItem_not_found, Exception) as e:
        command.state = COMMAND_ERROR
        command.result = str(e)
        command.time = time.time()

        logger.exception(str(e))

        try:
            commands.update(command)
            conn.close()
        except Exception:
            pass
        logger.info("Failed.")
        return

    try:
        command.state = COMMAND_DONE
        command.time = time.time()
        commands.update(command)
    except Exception as e:
        logger.exception(str(e))

    logger.info("Done.")
    logger.debug(" in %s s." % (str((time.time()*1000 - start_time)/1000),))


def configuration(config):
    class Configuration(PluginConfiguration):
        def __init__(self, config, *args, **kwargs):
            PluginConfiguration.__init__(self,
                                         config.configfile,
                                         *args,
                                         **kwargs)

            self.plugin_configuration = {
                'pg_ctl': None,
            }

            try:
                self.check_section(__name__)
            except ConfigurationError:
                return

            try:
                val = self.get(__name__, 'pg_ctl')
                for char in ['"', '\'']:
                    if val.startswith(char) and val.endswith(char):
                        val = val[1:-1]
                self.plugin_configuration['pg_ctl'] = val
            except NoOptionError:
                pass

    c = Configuration(config)
    return c.plugin_configuration



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from argparse import ArgumentParser, SUPPRESS as UNDEFINED_ARGUMENT
from sys import stdout
from getpass import getpass

from ..cli import cli
from ..usermgmt import hash_password
from ..errors import ConfigurationError, HTTPError, UserError
from ..usermgmt import get_user
from ..configuration import load_configuration
from ..types import T_PASSWORD, T_USERNAME
from ..tools import validate_parameters
from .agent import list_options_specs


def ask_password():
    raw_pass1 = getpass("Password: ")
    raw_pass2 = getpass("Retype password: ")
    if raw_pass1 != raw_pass2:
        stdout.write("Sorry, passwords do not match.\n")
        return ask_password()
    try:
        password = raw_pass1
        validate_parameters({'password': password},
                            [('password', T_PASSWORD, False)])
    except HTTPError:
        stdout.write("Invalid password.\n")
        return ask_password()
    return password


def ask_username(config):
    stdout.write("Username: ".encode('utf-8'))
    raw_username = raw_input()
    try:
        get_user(config.temboard['users'], raw_username)
    except HTTPError:
        pass
    except ConfigurationError:
        pass
    else:
        stdout.write("User already exists.\n")
        return ask_username(config)
    try:
        username = raw_username
        validate_parameters({'username': username},
                            [('username', T_USERNAME, False)])
    except HTTPError:
        stdout.write("Invalid username.\n")
        return ask_username(config)
    return username


@cli
def main(argv, environ):
    parser = ArgumentParser(
        prog='temboard-agent-adduser',
        description="Add a new temboard-agent user.",
        argument_default=UNDEFINED_ARGUMENT,
    )
    args = parser.parse_args(argv)
    config = load_configuration(
        specs=list_options_specs(), args=args, environ=environ,
    )

    # Load configuration from the configuration file.
    username = ask_username(config)
    password = ask_password()
    hash_ = hash_password(username, password).decode('utf-8')
    try:
        with open(config.temboard['users'], 'a') as fd:
            fd.write("%s:%s\n" % (username, hash_))
    except IOError as e:
        raise UserError(str(e))
    else:
        stdout.write("Done.\n")


if __name__ == '__main__':
    main()



import base64
import time
from hashlib import sha256, sha512
from os import urandom
from binascii import hexlify

from temboardagent.errors import ConfigurationError, HTTPError


def hash_password(username, password):
    """
    Hash a password with the following formula:
        sha512(password + sha512(username))
    """
    bytes_password = hexlify(password.encode('utf-8'))
    bytes_sha_username = sha512(username.encode('utf-8')).digest()
    hpasswd = sha512(bytes_password + bytes_sha_username).digest()
    return base64.b64encode(hpasswd)


def read_password_file(filepath):
    """
    Read and return the user/password file.
    """
    try:
        with open(filepath, 'r') as fd:
            lines = fd.readlines()
            fd.close()
            return lines
    except IOError:
        raise ConfigurationError("Can't open password file for reading.")


def get_user(filepath, username):
    """
    Get a user/passwd form the file.
    """
    for line in read_password_file(filepath):
        (l_username, l_hpasswd) = line.strip().split(':')
        if username == l_username:
            return (l_username, l_hpasswd)
    raise HTTPError(404, 'Invalid username/password.')


def auth_user(filepath, username, password):
    """
    Hash and compair the given couple username/password.
    """
    (l_username, l_hpasswd) = get_user(filepath, username)
    n_hpasswd = hash_password(username, password)
    if n_hpasswd.decode('utf-8') != l_hpasswd:
        raise HTTPError(404, 'Invalid username/password.')


def gen_sessionid(username):
    """
    Sessionid generator.
    """
    bytes_sha_username = sha512(username.encode('utf-8')).digest()
    bytes_rand = hexlify(urandom(32))
    bytes_time = hexlify(str(time.time()*1000).encode('utf-8'))
    return sha256(
                bytes_sha_username +
                bytes_rand +
                bytes_time
            ).hexdigest()



import json
import logging
import pprint

from monitoring.probes import Probe
from temboardagent.httpsclient import https_request


class OutputEncoder(json.JSONEncoder):
    """Tell json that probe objects should be encoded as strings using their
    name.
    """
    def default(self, obj):
        if isinstance(obj, Probe):
            return repr(obj)
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def send_output(ca_cert_file, url, key, j_output):
    """Send data to the target URL."""

    logging.debug("Sending output to %s", url)
    logging.debug("Output is:" + pprint.pformat(j_output))
    (code, res, _) = https_request(
        ca_cert_file,
        method='POST',
        url=url,
        headers={
            "Content-type": "application/json",
            "X-Key": key
        },
        data=json.loads(j_output)
    )
    logging.debug("Server replied with status %s", code)


def remove_passwords(instances):
    clean_instances = []
    for instance in instances:
        clean_instance = {}
        for k in instance.keys():
            if k != 'password':
                clean_instance[k] = instance[k]
        clean_instances.append(clean_instance)
    return clean_instances



from __future__ import unicode_literals

from argparse import ArgumentParser, SUPPRESS as UNDEFINED_ARGUMENT
import os
from sys import stdout, stderr
from getpass import getpass
import re
import urllib2
import json
import logging

from ..cli import cli, define_common_arguments
from ..errors import (
    ConfigurationError,
    HTTPError,
)
from ..types import T_PASSWORD, T_USERNAME
from ..tools import validate_parameters
from ..httpsclient import https_request
from ..configuration import load_configuration, setup_logging
from agent import list_options_specs

logger = logging.getLogger(__name__)

def ask_password():
    try:
        raw_pass = os.environ['TEMBOARD_UI_PASSWORD']
    except KeyError:
        raw_pass = getpass(" Password: ")

    try:
        password = raw_pass
        validate_parameters({'password': password},
                            [('password', T_PASSWORD, False)])
    except HTTPError:
        stdout.write("Invalid password.\n")
        return ask_password()
    return password


def ask_username():
    try:
        raw_username = os.environ['TEMBOARD_UI_USER']
    except KeyError:
        raw_username = raw_input(" Username: ")

    try:
        username = raw_username
        validate_parameters({'username': username},
                            [('username', T_USERNAME, False)])
    except HTTPError:
        stdout.write("Invalid username.\n")
        return ask_username()
    return username


def define_arguments(parser):
    parser.add_argument(
        '-?', '--help',
        action='help',
        help='show this help message and exit')

    define_common_arguments(parser)

    parser.add_argument(
        '-h', '--host',
        dest='host',
        help="Agent address. Default: %(default)s",
        default='localhost'
    )
    parser.add_argument(
        '-p', '--port',
        dest='port',
        help="Agent listening TCP port. Default: %(default)s",
        default='2345',
    )
    parser.add_argument(
        '-g', '--groups',
        dest='groups',
        help="Instance groups list, comma separated. Default: %(default)s",
        default=None,
    )
    parser.add_argument(
        'ui_address',
        metavar='TEMBOARD-UI-ADDRESS',
        help="temBoard UI address to register to.",
    )


@cli
def main(argv, environ):
    parser = ArgumentParser(
        prog='temboard-agent-register',
        description=(
            "Register a couple PostgreSQL instance/agent "
            "to a Temboard UI."
        ),
        add_help=False,
        argument_default=UNDEFINED_ARGUMENT,
    )
    define_arguments(parser)

    args = parser.parse_args(argv)

    setup_logging(level='ERROR')

    # Loading agent configuration file.
    config = load_configuration(
        specs=list_options_specs(),
        args=args, environ=environ,
    )
    config.setup_logging()

    # Load configuration from the configuration file.
    try:
        # Getting system/instance informations using agent's discovering API
        print("Getting system & PostgreSQL informations from the agent "
              "(https://%s:%s/discover) ..." % (args.host, args.port))
        (code, content, cookies) = https_request(
                None,
                'GET',
                "https://%s:%s/discover" % (args.host, args.port),
                headers={
                    "Content-type": "application/json"
                })
        infos = json.loads(content)

        print("Login at %s ..." % (args.ui_address))
        username = ask_username()
        password = ask_password()
        (code, content, cookies) = https_request(
                None,
                'POST',
                "%s/json/login" % (args.ui_address),
                headers={
                    "Content-type": "application/json"
                },
                data={'username': username, 'password': password})
        temboard_cookie = None
        for cookie in cookies.split("\n"):
            cookie_content = cookie.split(";")[0]
            if re.match(r'^temboard=.*$', cookie_content):
                temboard_cookie = cookie_content
                continue

        if args.groups:
            groups = [g for g in args.groups.split(',')]
        else:
            groups = None

        # POSTing new instance
        print("Registering instance/agent to %s ..." % (args.ui_address))
        (code, content, cookies) = https_request(
                None,
                'POST',
                "%s/json/register/instance" % (args.ui_address),
                headers={
                    "Content-type": "application/json",
                    "Cookie": temboard_cookie
                },
                data={
                    'hostname': infos['hostname'],
                    'agent_key': config.temboard['key'],
                    'agent_address': args.host,
                    'agent_port': str(config.temboard['port']),
                    'cpu': infos['cpu'],
                    'memory_size': infos['memory_size'],
                    'pg_port': infos['pg_port'],
                    'pg_data': infos['pg_data'],
                    'pg_version': infos['pg_version'],
                    'plugins': infos['plugins'],
                    'groups': groups
                })
        if code != 200:
            raise HTTPError(code, content)
        print("Done.")
    except (ConfigurationError, HTTPError, Exception) as e:
        if isinstance(e, urllib2.HTTPError):
            err = json.loads(e.read())
            stderr.write("FATAL: %s\n" % err['error'])
        else:
            stderr.write("FATAL: %s\n" % str(e))
        return 1

    return 0


if __name__ == '__main__':  # pragma: no cover
    main()



from subprocess import Popen, PIPE, CalledProcessError, check_call
from shlex import split as shlex_split
import errno


def exec_command(command_args, **kwargs):
    """
    Execute a system command with Popen.
    """
    kwargs.setdefault("stdout", PIPE)
    kwargs.setdefault("stderr", PIPE)
    kwargs.setdefault("stdin", PIPE)
    kwargs.setdefault("close_fds", True)
    try:
        process = Popen(command_args, **kwargs)
    except OSError as err:
        return (err.errno, None, err.strerror)

    (stdout, stderrout) = process.communicate()

    return (process.returncode, stdout, stderrout)


def exec_script(script_args, **kwargs):
    """
    Execute an external script.
    """
    kwargs.setdefault("stderr", PIPE)
    kwargs.setdefault("stdout", PIPE)
    try:
        check_call(script_args, **kwargs)
    except CalledProcessError as err:
        return (err.returncode, None, err.output)
    except IOError as err:
        if err.errno == errno.EPIPE:
            pass

    return (0, None, None)


def oneline_cmd_to_array(command_line):
    """
    Split a command line using shlex module.
    """
    return shlex_split(command_line)



import logging
import time
import signal
import json
import sys

from os import getpid
try:
    from configparser import NoOptionError
except ImportError:
    from ConfigParser import NoOptionError

from temboardagent.api_wrapper import (
    api_function_wrapper,
    api_function_wrapper_pg,
)
from temboardagent.spc import connector, error
from temboardagent.routing import add_route, add_worker
from temboardagent.configuration import (
    PluginConfiguration,
    ConfigurationError,
)
from temboardagent.sharedmemory import Command
from temboardagent.tools import hash_id
from temboardagent.errors import (
    SharedItem_exists,
    SharedItem_no_free_slot_left,
)
from temboardagent.workers import COMMAND_START
from temboardagent.queue import Queue, Message
import dashboard.metrics as metrics


logger = logging.getLogger(__name__)


@add_route('GET', '/dashboard')
def dashboard(http_context,
              queue_in=None,
              config=None,
              sessions=None,
              commands=None):
    """
Get the whole last data set used to render dashboard view. Data have been collected async.

.. sourcecode:: http

    GET /dashboard HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:57:52 GMT
    Content-type: application/json

    {
        "active_backends":
        {
            "nb": 1,
            "time": 1429617751.29224
        },
        "loadaverage": 0.28,
        "os_version": "Linux 3.16.0-34-generic x86_64",
        "pg_version": "9.4.1",
        "n_cpu": "4",
        "hitratio": 98.0,
        "databases":
        {
            "total_size": "1242 MB",
            "time": "14:02",
            "databases": 4,
            "total_commit": 16728291,
            "total_rollback": 873
        },
        "memory": {
            "total": 3950660,
            "active": 46.9,
            "cached": 20.2,
            "free": 32.9
        },
        "hostname": "neptune",
        "cpu":
        {
            "iowait": 0.0,
            "idle": 97.5,
            "steal": 0.0,
            "user": 2.5,
            "system": 0.0
        },
        "buffers":
        {
            "nb": 348247,
            "time": 1429617751.276508
        }
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Parameter 'X-Session' is malformed."}


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_metrics_queue')


@add_route('GET', '/dashboard/live')
def dashboard_live(http_context,
                   queue_in=None,
                   config=None,
                   sessions=None,
                   commands=None):
    """
Synchronous version of ``/dashboard``. Please refer to ``/dashboard`` API documentation for details.
    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_metrics')


@add_route('GET', '/dashboard/history')
def dashboard_history(http_context,
                      queue_in=None,
                      config=None,
                      sessions=None,
                      commands=None):
    """
Get the last ``n`` sets of dashboard data. ``n`` is defined by parameter ``history_length`` from the ``dashboard`` section of configuration file. Default value is ``20``.

.. sourcecode:: http

    GET /dashboard/history HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 15:56:56 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    [
        {
            "active_backends":
            {
                "nb": 1,
                "time": 1492703660.798522
            },
            "databases":
            {
                "total_rollback": 1081,
                "total_size": "158 MB",
                "timestamp": 1492703660.913077,
                "time": "17:54",
                "total_commit": 2825374,
                "databases": 6
            },
            "hostname": "poseidon.home.priv",
            "pg_version": "PostgreSQL 9.5.5 on x86_64-pc-linux-gnu, compiled by x86_64-pc-linux-gnu-gcc (Gentoo 4.9.4 p1.0, pie-0.6.4) 4.9.4, 64-bit",
            "memory":
            {
                "active": 51.0,
                "cached": 29.5,
                "total": 8082124,
                "free": 19.5
            },
            "cpu":
            {
                "iowait": 0.0,
                "idle": 100.0,
                "steal": 0.0,
                "user": 0.0,
                "system": 0.0
            },
            "os_version": "Linux 4.9.6-gentoo-r1",
            "loadaverage": 0.18,
            "hitratio": 99.0,
            "pg_uptime": "01:50:31.573788",
            "pg_port": "5432",
            "n_cpu": 4,
            "pg_data": "/var/lib/postgresql/9.5/data",
            "buffers":
            {
                "nb": 27670,
                "time": 1492703660.784254
            }
        }
    ]

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_history_metrics_queue')


@add_route('GET', '/dashboard/buffers')
def dashboard_buffers(http_context,
                      queue_in=None,
                      config=None,
                      sessions=None,
                      commands=None):
    """
Get the number of buffers allocated by PostgreSQL ``background writer`` process.

.. sourcecode:: http

    GET /dashboard/buffers HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:09:58 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {"buffers": {"nb": 27696, "time": 1492704598.784161}}

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_buffers')


@add_route('GET', '/dashboard/hitratio')
def dashboard_hitratio(http_context,
                       queue_in=None,
                       config=None,
                       sessions=None,
                       commands=None):
    """
Get PostgreSQL global cache hit ratio.

.. sourcecode:: http

    GET /dashboard/hitratio HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:28:33 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {"hitratio": 99.0}

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_hitratio')


@add_route('GET', '/dashboard/active_backends')
def dashboard_active_backends(http_context,
                              queue_in=None,
                              config=None,
                              sessions=None,
                              commands=None):
    """
Get the total number of active backends.

.. sourcecode:: http

    GET /dashboard/active_backends HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:35:55 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "active_backends":
        {
            "nb": 1,
            "time": 1492706155.986045
        }
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_active_backends')


@add_route('GET', '/dashboard/cpu')
def dashboard_cpu(http_context,
                  queue_in=None,
                  config=None,
                  sessions=None,
                  commands=None):
    """
Get CPU usage.

.. sourcecode:: http

    GET /dashboard/cpu HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:40:46 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "cpu":
        {
            "iowait": 0.0,
            "idle": 100.0,
            "steal": 0.0,
            "user": 0.0,
            "system": 0.0
        }
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_cpu_usage')


@add_route('GET', '/dashboard/loadaverage')
def dashboard_loadaverage(http_context,
                          queue_in=None,
                          config=None,
                          sessions=None,
                          commands=None):
    """
System loadaverage.

.. sourcecode:: http

    GET /dashboard/loadaverage HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:44:04 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "loadaverage": 0.06
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.
    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_loadaverage')


@add_route('GET', '/dashboard/memory')
def dashboard_memory(http_context,
                     queue_in=None,
                     config=None,
                     sessions=None,
                     commands=None):
    """
Memory usage.

.. sourcecode:: http

    GET /dashboard/memory HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:46:39 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "memory":
        {
            "active": 50.1,
            "cached": 29.5,
            "total": 8082124,
            "free": 20.4
        }
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_memory_usage')


@add_route('GET', '/dashboard/hostname')
def dashboard_hostname(http_context,
                       queue_in=None,
                       config=None,
                       sessions=None,
                       commands=None):
    """
Machine hostname.

.. sourcecode:: http

    GET /dashboard/hostname HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:48:49 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "hostname": "poseidon.home.priv"
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_hostname')


@add_route('GET', '/dashboard/os_version')
def dashboard_os_version(http_context,
                         queue_in=None,
                         config=None,
                         sessions=None,
                         commands=None):
    """
Operating system version.

.. sourcecode:: http

    GET /dashboard/os_version HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:55:44 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "os_version": "Linux 4.9.6-gentoo-r1"
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_os_version')


@add_route('GET', '/dashboard/pg_version')
def dashboard_pg_version(http_context,
                         queue_in=None,
                         config=None,
                         sessions=None,
                         commands=None):
    """
Get PostgreSQL server version.

.. sourcecode:: http

    GET /dashboard/pg_version HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 16:59:26 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "pg_version": "PostgreSQL 9.5.5 on x86_64-pc-linux-gnu, compiled by x86_64-pc-linux-gnu-gcc (Gentoo 4.9.4 p1.0, pie-0.6.4) 4.9.4, 64-bit"
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_pg_version')


@add_route('GET', '/dashboard/n_cpu')
def dashboard_n_cpu(http_context,
                    queue_in=None,
                    config=None,
                    sessions=None,
                    commands=None):
    """
Number of CPU.

.. sourcecode:: http

    GET /dashboard/n_cpu HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 17:03:55 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "n_cpu": 4
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                metrics,
                                'get_n_cpu')


@add_route('GET', '/dashboard/databases')
def dashboard_databases(http_context,
                        queue_in=None,
                        config=None,
                        sessions=None,
                        commands=None):
    """
PostgreSQL cluster size & number of databases.

.. sourcecode:: http

    GET /dashboard/databases HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 17:08:59 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "databases":
        {
            "total_rollback": 1087,
            "total_size": "159 MB",
            "timestamp": 1492708139.981268,
            "databases": 6,
            "total_commit": 2848707,
            "time": "19:08"
        }
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_databases')


@add_route('GET', '/dashboard/info')
def dashboard_info(http_context,
                   queue_in=None,
                   config=None,
                   sessions=None,
                   commands=None):
    """
Get a bunch of global informations about system and PostgreSQL.

.. sourcecode:: http

    GET /dashboard/info HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Thu, 20 Apr 2017 17:17:57 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "hostname": "poseidon.home.priv",
        "os_version": "Linux 4.9.6-gentoo-r1",
        "pg_port": "5432",
        "pg_uptime": "03:14:08.029574",
        "pg_version": "PostgreSQL 9.5.5 on x86_64-pc-linux-gnu, compiled by x86_64-pc-linux-gnu-gcc (Gentoo 4.9.4 p1.0, pie-0.6.4) 4.9.4, 64-bit",
        "pg_data": "/var/lib/postgresql/9.5/data"
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.

    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   metrics,
                                   'get_info')


def dashboard_worker_sigterm_handler(signum, frame):
    sys.exit(1)


@add_worker(b'dashboard_collector')
def dashboard_collector_worker(commands, command, config):
    try:
        signal.signal(signal.SIGTERM, dashboard_worker_sigterm_handler)
        start_time = time.time() * 1000
        logger.debug("Starting with pid=%s" % (getpid()))
        logger.debug("commandid=%s" % (command.commandid))
        command.state = COMMAND_START
        command.time = time.time()
        command.pid = getpid()
        commands.update(command)

        conn = connector(
            host=config.postgresql['host'],
            port=config.postgresql['port'],
            user=config.postgresql['user'],
            password=config.postgresql['password'],
            database=config.postgresql['dbname']
        )
        conn.connect()
        db_metrics = metrics.get_metrics(conn, config)
        # We don't want to store notifications in the history.
        db_metrics.pop('notifications', None)

        conn.close()
        q = Queue('%s/dashboard.q' % (config.temboard['home']),
                  max_length=(config.plugins['dashboard']['history_length']+1),
                  overflow_mode='slide'
                  )
        q.push(Message(content=json.dumps(db_metrics)))
        logger.debug("Duration: %s." % (str(time.time() * 1000 - start_time)))
        logger.debug("Done.")
    except (error, Exception) as e:
        logger.exception(str(e))
        logger.debug("Failed.")
        try:
            conn.close()
        except Exception:
            pass
        sys.exit(1)


def scheduler(queue_in, config, commands):
    worker = b'dashboard_collector'
    parameters = ''
    # Check command uniqueness.
    try:
        commands.check_uniqueness(worker, parameters)
    except SharedItem_exists:
        return

    cid = hash_id(worker)
    command = Command(cid.encode('utf-8'),
                      time.time(),
                      0,
                      worker,
                      parameters,
                      0,
                      u''
                      )
    try:
        commands.add(command)
        # Put the Command in the command queue
        queue_in.put(command)
        return
    except SharedItem_no_free_slot_left:
        return


def configuration(config):
    class Configuration(PluginConfiguration):
        def __init__(self, config, *args, **kwargs):
            PluginConfiguration.__init__(self,
                                         config.configfile,
                                         *args,
                                         **kwargs)

            self.plugin_configuration = {
                'scheduler_interval': 2,
                'history_length': 20
            }

            try:
                self.check_section(__name__)
            except ConfigurationError:
                return

            try:
                if not (self.getint(__name__, 'scheduler_interval') > 0 and
                        self.getint(__name__, 'scheduler_interval') < 86400):
                    raise ValueError()
                self.plugin_configuration['scheduler_interval'] = \
                    self.getint(__name__, 'scheduler_interval')
            except ValueError:
                logger.error("%s - configuration error: 'scheduler_interval' "
                             "must be an integer between 0 and 86400 in "
                             "section '%s' in %s." % (
                                 __name__,
                                 self.configfile,
                                 __name__
                                 )
                             )
            except NoOptionError:
                pass

            try:
                if not (self.getint(__name__, 'history_length') > 0 and
                        self.getint(__name__, 'history_length') < 300):
                    raise ValueError()
                self.plugin_configuration['history_length'] = \
                    self.getint(__name__, 'history_length')
            except ValueError:
                logger.error("%s - configuration error: 'histor_length' must "
                             "be an integer between 0 and 300 in section '%s'"
                             " in %s." % (
                                 __name__,
                                 self.configfile,
                                 __name__
                                 )
                             )
            except NoOptionError:
                pass

    c = Configuration(config)
    return c.plugin_configuration



import platform
import socket
import os
import re

from temboardagent.spc import connector
from temboardagent.tools import check_fqdn, which, to_bytes
from temboardagent.command import exec_command


class Inventory(object):
    def __init__(self):
        pass


class SysInfo(Inventory):
    def __init__(self):
        (self.os, self.os_release) = self._os_info()

    def _os_info(self):
        return (platform.system(), platform.release())

    def hostname(self, hostname=None):
        if not hostname:
            # Find the hostname by ourself.
            if self.os == 'Linux':
                hostname = self._hostname_linux()
            else:
                raise Exception("Unsupported OS.")
        if not check_fqdn(hostname):
            raise ValueError("Invalid FQDN: %s" % (hostname))
        return hostname

    def uname(self):
        return os.uname()

    def n_cpu(self):
        """
        Returns number of cpu using multiprocessinf.cpu_count().
        """
        from multiprocessing import cpu_count
        return cpu_count()

    def memory_size(self):
        if self.os == 'Linux':
            return self._mem_info_linux()['MemTotal']
        else:
            raise Exception("Unsupported OS.")

    def cpu_info(self):
        if self.os == 'Linux':
            return self._cpu_info_linux()
        else:
            raise Exception("Unsupported OS.")

    def mem_info(self):
        if self.os == 'Linux':
            return self._mem_info_linux()
        else:
            raise Exception("Unsupported OS.")

    def ip_addresses(self):
        if self.os == 'Linux':
            return self._ip_addresses_linux()
        else:
            raise Exception("Unsupported OS.")

    def file_systems(self):
        if self.os == 'Linux':
            return self._file_systems_linux()
        else:
            raise Exception("Unsupported OS.")

    def find_mount_point(self, path, mount_points):
        if self.os == 'Linux':
            return self._find_mount_point_linux(path, mount_points)
        else:
            raise Exception("Unsupported OS.")

    def os_flavor(self):
        if self.os == 'Linux':
            return self._os_flavor_linux()
        else:
            raise Exception("Unsupported OS.")

    def _hostname_linux(self):
        """
        Returns system hostname.
        """
        # Default value found using platform
        hostname = platform.node()
        try:
            # Try to get hostname (FQDN) using 'hostname -f'
            (rc, out, err) = exec_command([which('hostname'), '-f'])
            if rc == 0:
                hostname = out.encode('utf-8').strip()
        except Exception:
            try:
                # Try to get hostname (FQDN) using socket module
                (hostname, _, _) = socket.gethostbyaddr(socket.gethostname())
                hostname = hostname.strip()
            except Exception:
                pass
        return hostname

    def _cpu_info_linux(self):
        cpus = []
        # TODO: implement for other OSes
        if self.os != 'Linux':
            return {}
        current_cpu = {}
        with open('/proc/cpuinfo') as f:
            for line in f.read().split("\n"):
                if ':' not in line:
                    continue
                key, value = [part.strip() for part in line.split(':', 1)]
                if key == 'processor':
                    current_cpu = {}
                    cpus.append(current_cpu)
                if key in ['core id', 'cpu MHz', 'model name',
                           'cache size', 'processor']:
                    current_cpu[str(re.sub(r' +', '_', key)).lower()] = value
        if cpus:
            return {
                'cpus': cpus,
                'cpu_count': len(cpus)
            }
        return {}

    def _mem_info_linux(self):
        unit_re = re.compile(r'(\d+) ?(\wB)?')
        mem_values = {}
        with open('/proc/meminfo') as f:
            for line in f.read().split("\n"):
                if ':' not in line:
                    continue
                key, value = [part.strip() for part in line.split(':', 1)]
                size, unit = unit_re.match(value).groups()
                size = int(size)
                # convert everything to bytes, if a unit is specified
                if unit:
                    size = to_bytes(size, unit[:-1])
                mem_values[key] = size
        return mem_values

    def _ip_addresses_linux(self):
        """Find the host's IP addresses."""
        addrs = []
        try:
            ip = which('ip')
            (rc, out, err) = exec_command([ip, "addr", "show"])
            if rc == 0:
                for line in out.decode('utf8').splitlines():
                    m = re.match(r'^\s+inet ([\d\.]+)/\d+\s', line)
                    if m:
                        addrs.append(m.group(1))

                    m = re.match(r'^\sinet6 ([\dabcdef\:]+)/\d+ scope global',
                                 line)
                    if m:
                        addrs.append(m.group(1))
                return addrs
        except OSError:
            pass

        try:
            ifconfig = which('ifconfig', ['/sbin'])
            (rc, out, err) = exec_command([ifconfig, "-a"])
            if rc == 0:
                for line in out.splitlines():
                    m = re.match(r'^\s+inet (addr:)?([\d\.]+)\s', line)
                    if m:
                        addrs.append(m.group(2))

                    m = re.match(r'^\sinet6 (addr: )?([\dabcdef\:]+)(/\d+)? '
                                 '.+[Gg]lobal$', line)
                    if m:
                        addrs.append(m.group(2))
                return addrs
        except OSError:
            pass

        return addrs

    def _file_systems_linux(self):
        fs = []
        (rc, out, err) = exec_command([which('df'), '-k'])
        lines = out.splitlines()
        # Remove header
        del lines[0]
        dev = None
        for line in lines:
            cols = line.split()
            # Skip rootfs which is redundant on Debian
            if cols[0] == 'rootfs':
                continue

            # Output of df can be multiline when the name of the
            # device is too large
            if len(cols) == 1:
                dev = cols[0]
                continue
            if dev is not None:
                fs.append({
                    'mount_point': cols[4].decode('utf-8'),
                    'device': dev.decode('utf-8'),
                    'total': int(cols[0]) * 1024,
                    'used': int(cols[1]) * 1024
                })
                dev = None
            else:
                # Single line output from df
                fs.append({
                    'mount_point': cols[5].decode('utf-8'),
                    'device': cols[0].decode('utf-8'),
                    'total': int(cols[1]) * 1024,
                    'used': int(cols[2]) * 1024
                })
        return fs

    def mount_points(self):
        return [fs['mount_point'] for fs in self.file_systems()]

    def _find_mount_point_linux(self, path, mount_points):
        realpath = os.path.realpath(path)

        if not os.path.exists(realpath):
            return None

        # Get the parent dir when it is not a directory
        if not os.path.isdir(realpath):
            realpath = os.path.dirname(realpath)

        # Walk up parents directory
        while True:
            if realpath in mount_points or realpath == '/':
                return realpath

            realpath = os.path.dirname(realpath)

    def _os_flavor_linux(self):
        # Distribution
        os_flavor = "Unknown"
        if os.path.exists("/etc/redhat-release"):
            try:
                fd = open("/etc/redhat-release")
                os_flavor = fd.readline().strip()
                fd.close()
            except OSError:
                fd.close()
        elif os.path.exists("/etc/debian_version"):
            try:
                fd = open("/etc/debian_version")
                os_flavor = "Debian " + fd.readline().strip()
                fd.close()
            except OSError:
                fd.close()
        return os_flavor


class PgInfo(Inventory):
    def __init__(self, db_conn):
        if isinstance(db_conn, connector):
            self.db_conn = db_conn
        else:
            raise Exception("db_conn should be an SPC connector instance.")

    def setting(self, name):
        """
        Returns PostgreSQL setting value based on its name.
        """
        q = "SELECT setting FROM pg_settings WHERE name = %s"
        self.db_conn.execute(q, (name,))
        return list(self.db_conn.get_rows())[0]['setting']

    def version(self):
        """
        Returns a dict with PostgreSQL full & numeric version.
        """
        q = "SELECT version(), setting AS server FROM pg_settings WHERE " \
            "name = 'server_version'"
        self.db_conn.execute(q)
        return {
            'full': list(self.db_conn.get_rows())[0]['version'],
            'server': list(self.db_conn.get_rows())[0]['server'],
            'num': self.db_conn.get_pg_version()
        }

    def is_in_recovery(self):
        if self.db_conn.get_pg_version() >= 90000:
            self.db_conn.execute("SELECT pg_is_in_recovery() AS standby;")
            if list(self.db_conn.get_rows())[0]['standby'] == 't':
                return True
        return False

    def tablespaces(self, data_directory):
        # Grab the list of tablespaces
        if self.db_conn.get_pg_version() < 90200:
            self.db_conn.execute("SELECT spcname, spclocation, "
                                 "pg_tablespace_size(oid) AS size FROM "
                                 "pg_tablespace;")
        else:
            self.db_conn.execute("SELECT spcname, pg_tablespace_location(oid) "
                                 "AS spclocation, pg_tablespace_size(oid) AS "
                                 "size FROM pg_tablespace;")

        tablespaces = []
        sysinfo = SysInfo()
        fs = sysinfo.mount_points()
        for row in self.db_conn.get_rows():
            # when spclocation is empty, replace with data_directory
            if row['spclocation'] is None:
                path = data_directory
            else:
                path = row['spclocation']

            tablespaces.append({
                'spcname': row['spcname'],
                'path': path,
                'mount_point': sysinfo.find_mount_point(path, fs),
                'size': row['size']
            })

    def databases(self):
        self.db_conn.execute("""
        SELECT datname AS dbname, pg_encoding_to_char(encoding) AS encoding,
        pg_database_size(datname) AS size FROM pg_database WHERE datallowconn;
        """)

        dbs = {}
        for r in self.db_conn.get_rows():
            dbs[r['dbname']] = r
        return dbs



import logging

from temboardagent.api import check_sessionid
from temboardagent.errors import HTTPError
from temboardagent.spc import connector, error


logger = logging.getLogger(__name__)


def api_function_wrapper_pg(config, http_context, sessions, module,
                            function_name):
    """
    API function wrapper in charge of:
        - instanciating a new logger;
        - check the user session id;
        - start a new PostgreSQL connection;
        - call the function 'function_name' from 'module_name' module and
          return its result;
        - close PG connection.
    """
    logger.debug("Calling %s.%s()." % (module.__name__, function_name,))
    logger.debug(http_context)

    try:
        username = check_sessionid(http_context['headers'], sessions)
        http_context['username'] = username
        conn = connector(
            host=config.postgresql['host'],
            port=config.postgresql['port'],
            user=config.postgresql['user'],
            password=config.postgresql['password'],
            database=config.postgresql['dbname']
        )
        conn.connect()
        dm = getattr(module, function_name)(conn, config, http_context)
        conn.close()
        logger.debug("Done.")
        return dm

    except (error, Exception, HTTPError) as e:
        logger.exception(str(e))
        logger.debug("Failed.")
        try:
            conn.close()
        except Exception:
            pass
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")


def api_function_wrapper(config, http_context, sessions, module,
                         function_name):
    """
    API function wrapper in charge of:
        - instanciating a new logger;
        - check the user session id;
        - call the function 'function_name' from 'module_name' module and
          return its result;
    """
    logger.debug("Calling %s.%s()." % (module.__name__, function_name,))
    logger.debug(http_context)

    try:
        username = check_sessionid(http_context['headers'], sessions)
        http_context['username'] = username
        dm = getattr(module, function_name)(config, http_context)
        logger.debug("Done.")
        return dm

    except (Exception, HTTPError) as e:
        logger.exception(str(e))
        logger.debug("Failed.")
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")



# -*- coding: utf-8 -*-
"""
The inventory contains support functions to discover the features of the host
(ex: OS, CPU info, MemInfo)

Currently, only Linux is supported.
"""

import os
import logging
import pwd
from temboardagent.spc import (
    connector,
    error,
)
from temboardagent.inventory import (
    SysInfo,
    PgInfo,
)


def host_info(hostname_cfg):
    """Gather system information."""
    sinfo = SysInfo()
    _, _, _, _, arch = sinfo.uname()
    hostname = sinfo.hostname(hostname_cfg)

    hostinfo = {
        "hostname": hostname,
        "os": sinfo.os,
        "os_version": sinfo.os_release,
        "cpu_arch": arch
    }
    hostinfo.update(sinfo.cpu_info())
    hostinfo['memory_size'] = sinfo.mem_info()['MemTotal']
    hostinfo['ip_addresses'] = sinfo.ip_addresses()
    hostinfo['filesystems'] = sinfo.file_systems()
    hostinfo['os_flavor'] = sinfo.os_flavor()
    return hostinfo


def instance_info(conninfo, hostname):
    """Gather PostgreSQL instance information."""
    instance_info = {
        'hostname': hostname,
        'instance': conninfo['instance'],
        'local_name': conninfo.get('local_name', conninfo['instance']),
        'available': True,
        'host': conninfo['host'],
        'port': conninfo['port'],
        'user': conninfo['user'],
        'database': conninfo['database'],
        'password': conninfo['password']
    }

    # Try the connection
    conn = connector(conninfo['host'], conninfo['port'], conninfo['user'],
                     conninfo['password'], conninfo['database'])
    try:
        conn.connect()
        # Get PostgreSQL informations using PgInfo
        pginfo = PgInfo(conn)
        pgv = pginfo.version()
        # Gather the info while where are connected
        instance_info['version_num'] = pgv['num']
        instance_info['version'] = pgv['server']
        instance_info['data_directory'] = pginfo.setting('data_directory')

        # hot standby is available from 9.0
        instance_info['standby'] = pginfo.is_in_recovery()

        # Grab the list of tablespaces
        instance_info['tablespaces'] = pginfo.tablespaces(
            instance_info['data_directory'])

        # When the user has not given a dbnames list or '*' in the
        # configuration file, we must get the list of databases. Since
        # we have a working connection, let's do it now.
        dbs = pginfo.databases()
        instance_info['dbnames'] = []
        for db in conninfo['dbnames']:
            if db == '*':
                instance_info['dbnames'] = dbs.values()
                break
            if db in dbs.keys():
                instance_info['dbnames'].append(dbs[db])
        conn.close()

        # Now that we have the data_directory, find the owner
        try:
            statinfo = os.stat(instance_info['data_directory'])
            instance_info['sysuser'] = pwd.getpwuid(statinfo.st_uid).pw_name
        except OSError as e:
            logging.warning("Unable to get the owner of PGDATA: %s", str(e))
            instance_info['sysuser'] = None

    except error as e:
        logging.exception(str(e))
        logging.warning("Unable to gather information for cluster \"%s\"",
                        conninfo['instance'])
        instance_info['available'] = False

    return instance_info



from temboardagent.routing import add_route
from temboardagent.api_wrapper import api_function_wrapper_pg
import activity.functions as activity_functions


@add_route('GET', '/activity')
def get_activity(http_context,
                 queue_in=None,
                 config=None,
                 sessions=None,
                 commands=None):
    """
Get list of PostgreSQL backends.

**Example request**:

.. sourcecode:: http

    GET /activity HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:57:52 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "rows":
        [
            {
                "pid": 6285,
                "database": "postgres",
                "user": "postgres",
                "client": null,
                "cpu": 0.0,
                "memory": 0.13,
                "read_s": "0.00B",
                "write_s": "0.00B",
                "iow": "N",
                "wait": "N",
                "duration": "1.900",
                "state": "idle",
                "query": "SELECT 1;"
            }
        ]
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Parameter 'X-Session' is malformed."}


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   activity_functions,
                                   'get_activity')


@add_route('GET', '/activity/waiting')
def get_activity_waiting(http_context,
                         queue_in=None,
                         config=None,
                         sessions=None,
                         commands=None):
    """
Get list of PostgreSQL backends waiting for lock acquisition.

**Example request**:

.. sourcecode:: http

    GET /activity/waiting HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:57:52 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "rows":
        [
            {
                "pid": 13532,
                "database": "test",
                "user": "postgres",
                "cpu": 0.0,
                "memory": 0.16,
                "read_s": "0.00B",
                "write_s": "0.00B",
                "iow": "N",
                "relation": " ",
                "type": "transactionid",
                "mode": "ShareLock",
                "state": "active",
                "duration": 4.35,
                "query": "DELETE FROM t1 WHERE id = 1;"
            }
        ]
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.

    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   activity_functions,
                                   'get_activity_waiting')


@add_route('GET', '/activity/blocking')
def get_activity_blocking(http_context,
                          queue_in=None,
                          config=None,
                          sessions=None,
                          commands=None):
    """
Get list of PostgreSQL backends blocking other backends due to lock acquisition.

**Example request**:

.. sourcecode:: http

    GET /activity/blocking HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:57:52 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "rows":
        [
            {
                "pid": 13309,
                "database": "test",
                "user": "postgres",
                "cpu": 0.0,
                "memory": 0.2,
                "read_s": "0.00B",
                "write_s": "0.00B",
                "iow": "N",
                "relation": " ",
                "type": "transactionid",
                "mode": "ExclusiveLock",
                "state": "idle in transaction",
                "duration": 4126.98,
                "query": "UPDATE t1 SET id = 100000000 where id =1;"
            }
        ]
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   activity_functions,
                                   'get_activity_blocking')


@add_route('POST', '/activity/kill')
def post_activity_kill(http_context,
                       queue_in=None,
                       config=None,
                       sessions=None,
                       commands=None):
    """
Terminate (kill) a list of PostgreSQL backends.

**Example request**:

.. sourcecode:: http

    POST /activity/kill HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e
    Content-Type: application/json

    {
        "pids":
        [
            13309
        ]
    }

**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:57:52 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "backends":
        [
            {"pid": 13309, "killed": true},
        ]
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   activity_functions,
                                   'post_activity_kill')



# User Input Validator
#
# These functions provide stateless validation of user input, mainly CLI
# arguments and environment variables.
#
# On invalid input, a ValueError is raised. Other exceptions are considered a
# bug.
#
# A validator is idempotent. It must accepts what it returns.

import json
import logging
import os.path
import re
from distutils.util import strtobool
from logging.handlers import SysLogHandler

from .log import HANDLERS as LOG_METHODS


_address_re = re.compile(
    r'(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d)'
    r'(\.(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d'
    r')){3}$'
)


def address(raw):
    if not _address_re.match(raw):
        raise ValueError('invalid address')
    return raw


def boolean(raw):
    if raw in (True, False):
        return raw

    return bool(strtobool(raw))


def dir_(raw):
    raw = os.path.realpath(raw)
    if not os.path.isdir(raw):
        raise ValueError('Not a directory')
    return raw


def file_(raw):
    raw = os.path.realpath(raw)
    if not os.path.exists(raw):
        raise ValueError('%s: File not found' % raw)
    return raw


_identifier_re = re.compile(r'^[a-zA-Z0-9]+$')


def jsonlist(raw):
    if hasattr(raw, 'lower'):
        raw = json.loads(raw)

    if not isinstance(raw, list):
        raise ValueError('not a list')

    raw = [str(e) for e in raw]
    for entry in raw:
        if not _identifier_re.match(entry):
            raise ValueError('%s is invalid' % entry)

    return raw


def port(raw):
    port = int(raw)

    if 0 > port or port > 65635:
        raise ValueError('Port out of range')

    return port


def loglevel(raw):
    raw = raw.upper()
    if raw not in logging._levelNames:
        raise ValueError('unkown log level')
    return raw


def logmethod(raw):
    if raw not in LOG_METHODS:
        raise ValueError('unkown method')
    return raw


def syslogfacility(raw):
    if raw not in SysLogHandler.facility_names:
        raise ValueError('unkown syslog facility')
    return raw


def writeabledir(raw):
    raw = dir_(raw)
    if not os.access(raw, os.W_OK):
        raise ValueError('Not writable')
    return raw



import logging
try:
    from logging.config import dictConfig
except ImportError:  # pragma: nocover
    from logutils.dictconfig import dictConfig

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import os.path

from temboardagent.errors import ConfigurationError
from .utils import DotDict
from .pluginsmgmt import load_plugins_configurations
from .log import generate_logging_config
from .errors import UserError
from . import validators


logger = logging.getLogger(__name__)


def setup_logging(**kw):
    logging_config = generate_logging_config(**kw)
    dictConfig(logging_config)


class PluginConfiguration(configparser.RawConfigParser):
    """
    Customized configuration parser for plugins.
    """
    def __init__(self, configfile, *args, **kwargs):
        configparser.RawConfigParser.__init__(self, *args, **kwargs)
        self.configfile = configfile
        self.confdir = os.path.dirname(self.configfile)

        try:
            with open(self.configfile) as fd:
                self.readfp(fd)
        except IOError:
            raise ConfigurationError("Configuration file %s can't be opened."
                                     % (self.configfile))
        except configparser.MissingSectionHeaderError:
            raise ConfigurationError("Configuration file does not contain "
                                     "section headers.")

    def check_section(self, section):
        if not self.has_section(section):
            raise ConfigurationError("Section '%s' not found in configuration "
                                     "file %s" % (section, self.configfile))

    def abspath(self, path):
        if path.startswith('/'):
            return path
        else:
            return os.path.realpath('/'.join([self.confdir, path]))

    def getfile(self, section, name):
        path = self.abspath(self.get(section, name))
        try:
            with open(path) as fd:
                fd.read()
        except Exception as e:
            logger.warn("Failed to open %s: %s", path, e)
            raise ConfigurationError("%s file can't be opened." % (path,))
        return path


# Here begin the new API
#
# The purpose of the new API is to merge args, file, environment and defaults
# safely, even when reloading.
#
# The API must be very simple, IoC-free. Implementation must be highly testable
# and tested.


class OptionSpec(object):
    REQUIRED = object()

    # Hold known name and default of an option.
    #
    # An option *must* be specified to follow the principle of *validated your
    # inputs*.
    #
    # Defining defaults here is agnostic from origin : argparse, environ,
    # ConfigParser, etc. The origin of configuration must not take care of
    # default nor validation.
    #
    # Set default to OptionSpec.REQUIRED to enforce user definition of option
    # value.
    #
    # Option value can be None, meaning it is undefined.

    def __init__(self, section, name, validator=None, default=None):
        self.section = section
        self.name = name
        self.default = default
        self.validator = validator

    def __repr__(self):
        return '<OptionSpec %s>' % (self,)

    def __str__(self):
        return '%s_%s' % (self.section, self.name)

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    @property
    def required(self):
        return self.default is self.REQUIRED

    def validate(self, value):
        if value.value is None:
            return value.value

        if not self.validator:
            return value.value

        try:
            return self.validator(value.value)
        except ValueError as e:
            logger.debug(
                "Invalid %s from %s: %.32s...", value.name, value.origin, e)
            raise


def load_configuration(specs=None, args=None, environ=os.environ):
    # Main entry point to load configuration.
    #
    # specs is a list or a flat dict of OptionSpecs
    #
    # argparser should **not** manage defaults. Use argparse.SUPPRESS as
    # argument_default to store only user defined arguments. MergeConfiguration
    # merge defaults after file and environ are loaded. Defaults from argparse
    # are considered user input and overrides file and environ.
    #
    # configfile **must** be store in dest `temboard_configfile` in args.
    #
    # Origin order: args > environ > file > defaults

    config = MergedConfiguration(specs)
    config.load(args=args, environ=environ)
    return config


class Value(object):
    # Hold an option value and its origin
    def __init__(self, name, value, origin):
        self.name = name
        self.value = value
        self.origin = origin

    def __repr__(self):
        return '<%(name)s=%(value)r %(origin)s>' % self.__dict__


def iter_args_values(args):
    # Walk args from argparse and yield values.
    if not args:
        return

    for k, v in args.__dict__.items():
        yield Value(k, v, 'args')


def iter_configparser_values(parser, filename='config'):
    for section in parser.sections():
        for name, value in parser.items(section):
            name = '%s_%s' % (section, name)
            yield Value(name, value, origin=filename)


def iter_environ_values(environ):
    prefix = 'TEMBOARD_'
    for k, v in environ.items():
        if not k.startswith(prefix):
            continue

        k = k.lower()
        v = v.decode('utf-8')

        # Yield the value with temboard prefix so we don't have to define
        # TEMBOARD_TEMBOARD_* to set a value in temboard section.
        yield Value(k, v, 'environ')
        yield Value(k[len(prefix):], v, 'environ')


def iter_defaults(specs):
    # Walk specs flat dict and yield default values.
    for spec in specs.values():
        yield Value(str(spec), spec.default, 'defaults')


class MergedConfiguration(DotDict):
    # Merge and holds configuration from args, files and more
    #
    # Origin order: args > environ > file > defaults

    def __init__(self, specs=None):
        # Spec is a flat dict of OptionSpec.
        specs = specs or {}
        if not isinstance(specs, dict):
            specs = dict((s, s) for s in specs)

        # Add required configfile option
        spec = OptionSpec(
            'temboard', 'configfile',
            default='/etc/temboard-agent/temboard-agent.conf',
            validator=validators.file_,
        )
        specs.setdefault(spec, spec)

        DotDict.__init__(self)
        self.__dict__['specs'] = specs
        self.__dict__['unvalidated_specs'] = specs.keys()
        self.loaded = False

    def add_values(self, values):
        # Search missing values in values and validate them.

        values = dict((v.name, v) for v in values)
        for name in self.unvalidated_specs[:]:
            try:
                value = values[name]
            except KeyError:
                continue

            spec = self.specs[name]
            value = spec.validate(value)
            section = self.setdefault(spec.section, {})
            section[spec.name] = value
            self.unvalidated_specs.remove(name)

    def check_required(self):
        for name in self.unvalidated_specs:
            spec = self.specs[name]
            if spec.required:
                msg = "Missing %s:%s configuration" % (spec.section, spec.name)
                raise ConfigurationError(msg)

    def load(self, args, environ):
        # Origins are loaded in order. First wins..
        #
        # Loading in this order avoid validating ignored values.

        try:
            self.add_values(iter_args_values(args))
            self.add_values(iter_environ_values(environ))
            # Loading default for configfile *before* loading file.
            spec = self.specs['temboard_configfile']
            self.add_values([Value(str(spec), spec.default, 'default')])
            self.load_file(self.temboard.configfile)
        except ValueError as e:
            raise UserError(str(e))

        self.check_required()
        self.add_values(iter_defaults(self.specs))
        self.plugins = load_plugins_configurations(self)
        self.loaded = True

    def load_file(self, filename):
        parser = configparser.RawConfigParser()
        logger.info('Reading %s.', filename)
        try:
            with open(filename, 'ro') as fp:
                parser.readfp(fp)
        except IOError as e:
            raise ValueError(str(e))

        oldpwd = os.getcwd()
        os.chdir(os.path.dirname(filename))
        self.add_values(iter_configparser_values(parser, filename))
        try:
            os.chdir(oldpwd)
        except OSError as e:
            logger.debug("Can't move back to %s: %s", oldpwd, e)

        # Compat with legacy
        self.configfile = self.temboard.configfile

    def reload(self):
        # Reread file config.

        assert self.loaded, "Can't reload unloaded configuration."
        old_plugins = self.temboard.plugins

        try:
            self.load_file(self.temboard.configfile)
        except Exception as e:
            raise ConfigurationError(str(e))

        # Prevent any change on plugins list.
        self.temboard.plugins = old_plugins
        # Now reload plugins configurations
        self.plugins = load_plugins_configurations(self)
        return self

    def setup_logging(self):
        # Just to save one import for code reloading config.
        setup_logging(**self.logging)



import logging.config
from logging.handlers import SysLogHandler
import sys


class ColoredStreamHandler(logging.StreamHandler):

    _color_map = {
        logging.DEBUG: '37',
        logging.INFO: '1;39',
        logging.WARN: '96',
        logging.ERROR: '91',
        logging.CRITICAL: '1;91',
    }

    def format(self, record):
        lines = logging.StreamHandler.format(self, record)
        color = self._color_map.get(record.levelno, '39')
        lines = ''.join([
            '\033[0;%sm%s\033[0m' % (color, line)
            for line in lines.splitlines(True)
        ])
        return lines


class MultilineFormatter(logging.Formatter):
    def format(self, record):
        s = logging.Formatter.format(self, record)
        if '\n' not in s:
            return s

        lines = s.splitlines()
        d = record.__dict__.copy()
        for i, line in enumerate(lines[1:]):
            record.message = line
            lines[1+i] = self._fmt % record.__dict__
        record.__dict__ = d

        return '\n'.join(lines)


class LastnameFilter(logging.Filter):
    def filter(self, record):
        record.lastname = record.name
        if record.name.startswith('temboardagent.'):
            _, record.lastname = record.name.rsplit('.', 1)
        # Always log, we are just enriching records.
        return 1


HANDLERS = {
    'file': {
        '()': 'logging.FileHandler',
        'mode': 'a',
        'formatter': 'dated_syslog',
    },
    'syslog': {
        '()': 'logging.handlers.SysLogHandler',
        'formatter': 'syslog',
    },
    'stderr': {
        '()': 'logging.StreamHandler',
        'formatter': 'minimal',
    },
}


def generate_logging_config(
        level='DEBUG', destination=None, facility='local0',
        method='stderr', **kw):

    facility = SysLogHandler.facility_names[facility]
    HANDLERS['syslog']['facility'] = facility
    HANDLERS['syslog']['address'] = destination
    HANDLERS['file']['filename'] = destination

    fmt = 'verbose' if level == 'DEBUG' else 'minimal'
    HANDLERS['stderr']['formatter'] = fmt
    if sys.stderr.isatty():
        HANDLERS['stderr']['()'] = __name__ + '.ColoredStreamHandler'

    minimal_fmt = '%(levelname)5.5s: %(message)s'
    verbose_fmt = '%(asctime)s [%(lastname)-16.16s] ' + minimal_fmt
    syslog_fmt = (
        "temboard-agent[%(process)d]: "
        "[%(lastname)s] %(levelname)s: %(message)s"
    )

    logging_config = {
        'version': 1,
        'disable_existing_loggers': False,
        'filters': {
            'lastname': {
                '()': __name__ + '.LastnameFilter',
            }
        },
        'formatters': {
            'dated_syslog': {
                '()': __name__ + '.MultilineFormatter',
                'format': '%(asctime)s ' + syslog_fmt,
            },
            'minimal': {
                '()': __name__ + '.MultilineFormatter',
                'format': minimal_fmt,
            },
            'syslog': {
                '()': __name__ + '.MultilineFormatter',
                'format': syslog_fmt,
            },
            'verbose': {
                '()': __name__ + '.MultilineFormatter',
                'format': verbose_fmt,
            },
        },
        'handlers': {
            'configured': dict(
                HANDLERS[method],
                filters=['lastname'],
            ),
        },
        'root': {
            'level': 'INFO',
            # Avoid instanciate all handlers, especially syslog which open
            # /dev/log
            'handlers': ['configured'],
        },
        'loggers': {
            'temboardagent': {
                'level': level,
            },
            'temboard-agent': {
                'level': level,
            },
        },
    }
    return logging_config



class UserError(Exception):
    def __init__(self, message, retcode=1):
        super(UserError, self).__init__(message)
        self.retcode = retcode


class HTTPError(Exception):
    """ HTTP server errors """
    def __init__(self, code, message):
        Exception.__init__(self, message)
        self.code = code
        self.message = {'error': str(message)}


class ConfigurationError(UserError):
    pass


class SharedItem_not_found(Exception):
    def __init__(self,):
        pass


class SharedItem_exists(Exception):
    def __init__(self,):
        pass


class SharedItem_bad_type_size(Exception):
    def __init__(self,):
        pass


class SharedItem_no_free_slot_left(Exception):
    def __init__(self,):
        pass


class NotificationError(Exception):
    """ Notification errors """
    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = str(message)
